from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig
import time

logger = getLogger('GangaThread')


class GangaThreadPool(object):

    _attributes = ('SHUTDOWN_TIMEOUT')

    # GangaThreadPool singleton instance
    _instance = None

    class SingletonHelper(object):

        def __call__(self, *args, **kw):

            if GangaThreadPool._instance is None:
                object = GangaThreadPool()
                GangaThreadPool._instance = object

            return GangaThreadPool._instance

    getInstance = SingletonHelper()
    shutdown_policy = 'interactive'

    def __init__(self):

        if not GangaThreadPool._instance == None:
            raise RuntimeError(
                'Only one instance of GangaThreadPool is allowed!')

        GangaThreadPool._instance = self

        self.SHUTDOWN_TIMEOUT = 1
        self.__threads = []

    def addServiceThread(self, t):
        #logger.debug('service thread "%s" added to the GangaThreadPool', t.getName())
        ##   HERE TO AVOID AN IMPORT ERROR!
        from Ganga.Core.GangaThread.MTRunner import DuplicateDataItemError
        try:
            self.__threads.append(t)
        except DuplicateDataItemError as e:
            self.logger.debug("%s" % e)

    def delServiceThread(self, t):
        #logger.debug('service thread "%s" deleted from the GangaThreadPool', t.getName())
        try:
            if t in self.__threads:
                self.__threads.remove(t)
        except ValueError as e:
            logger.debug("%s" % e)

    def shutdown(self):
        """Shutdown the Ganga session.

        @param should_wait_cb: A callback function with the following signature
            should_wait_cb(total_time, critical_thread_ids, non_critical_thread_ids)
            where
                total_time is the time in seconds since shutdown started
                critical_thread_ids is a list of alive critical thread names
                non_critical_thread_ids is a list of alive non-critical threads names.
            and
                return value is evaluated as a boolean. 

        A shutdown thread is started that calls stop() on each GangaThread and
        waits for them all to die. A loop waits for the shutdown thread to
        die, periodically calling the should_wait_cb function to ask if it
        should continue to wait or shutdown anyway.

        """

        try:
            self._really_shutdown()
        except Exception as err:
            from Ganga.Utility.logging import getLogger
            logger = getLogger('GangaThread')
            logger.error("Error shutting down thread Pool!")
            logger.error("\n%s" % err)
        return

    def _really_shutdown(self):

        from Ganga.Utility.logging import getLogger
        logger = getLogger('GangaThread')

        logger.debug('shutting down GangaThreadPool with timeout %d sec' % self.SHUTDOWN_TIMEOUT)

        # run shutdown thread in background
        import threading
        shutdown_thread = threading.Thread(target=self.__do_shutdown__, args=(self.__threads,), name='GANGA_Update_Thread_shutdown')
        shutdown_thread.setDaemon(True)
        shutdown_thread.start()

        t_start = time.time()

        # Note that this is negative for the logic below to work
        t_last = -t_start

        def __cnt_alive_threads__(_all_threads):
            num_alive_threads = 0
            for t in _all_threads:
                if t.isAlive():
                    num_alive_threads += 1
            return num_alive_threads

        # wait for the background shutdown thread to finish
        config = getConfig('PollThread')
        while shutdown_thread.isAlive():
            logger.debug('Waiting for max %d seconds for threads to finish' % self.SHUTDOWN_TIMEOUT)
            logger.debug('There are %d alive background threads' % __cnt_alive_threads__(self.__threads))
            logger.debug('%s' % self.__alive_critical_thread_ids())
            logger.debug('%s' % self.__alive_non_critical_thread_ids())
            shutdown_thread.join(self.SHUTDOWN_TIMEOUT)

            # Have all the threads finished?
            if not shutdown_thread.isAlive():
                logger.debug('GangaThreadPool shutdown properly')
                break

            # Nope, so decide what we should do based on shutdown policy
            total_time = time.time() - t_start
            critical_thread_ids = self.__alive_critical_thread_ids()
            non_critical_thread_ids = self.__alive_non_critical_thread_ids()

            if GangaThreadPool.shutdown_policy == 'batch':
                # we have batch shutdown policy so wait until PollThread.forced_shutdown_timeout for critical threads
                # and PollThread.forced_shutdown_first_prompt_time for non-critical threads
                if critical_thread_ids:
                    if total_time > config['forced_shutdown_timeout']:
                        getLogger().warning('Shutdown was forced after waiting for %d seconds for background '
                                            'activities to finish (monitoring, output download, etc). This may '
                                            'result in some jobs being corrupted.', total_time)
                        break

                elif non_critical_thread_ids:
                    if total_time > config['forced_shutdown_first_prompt_time']:
                        break
                else:
                    break
            else:
                # we have interactive shutdown policy so for critical threads, wait until
                # PollThread.forced_shutdown_first_prompt_time before asking every
                # PollThread.forced_shutdown_prompt_time before asking to force exit. For non-critical threads wait
                # for PollThread.forced_shutdown_first_prompt_time before forcing the exit.
                if critical_thread_ids:
                    if ((t_last < 0 and time.time() + t_last > config['forced_shutdown_first_prompt_time']) or
                            (t_last > 0 and time.time() - t_last > config['forced_shutdown_prompt_time'])):
                        msg = 'Job status update or output download still in progress (shutdown not completed ' \
                              'after %d seconds). %d background thread(s) still running: %s. Do you want to force ' \
                              'the exit (y/[n])? ' % (total_time, len(critical_thread_ids), critical_thread_ids)
                        resp = raw_input(msg)
                        t_last = time.time()
                        if resp.lower() == 'y':
                            break
                elif non_critical_thread_ids:
                    if total_time > config['forced_shutdown_first_prompt_time']:
                        break
                else:
                    break

        # log warning message if critical thread still alive
        critical_thread_ids = self.__alive_critical_thread_ids()
        if critical_thread_ids:
            logger.debug('GangaThreadPool shutdown anyway after %d sec.' % (time.time() - t_start))
            logger.warning('Shutdown forced. %d background thread(s) still running: %s', len(critical_thread_ids), critical_thread_ids)

        # log debug message if critical thread still alive
        non_critical_thread_ids = self.__alive_non_critical_thread_ids()
        if non_critical_thread_ids:
            logger.debug('GangaThreadPool shutdown anyway after %d sec.' % (time.time() - t_start))
            logger.debug('Shutdown forced. %d non-critical background thread(s) still running: %s', len(non_critical_thread_ids), non_critical_thread_ids)

        # set singleton instance to None
        self._instance = None
        for i in self.__threads:
            del i
        self.__threads = []

    def __alive_critical_thread_ids(self):
        """Return a list of alive critical thread names."""
        return [t.gangaName for t in self.__threads if t.isAlive() and t.isCritical()]

    def __alive_non_critical_thread_ids(self):
        """Return a list of alive non-critical thread names."""
        return [t.gangaName for t in self.__threads if t.isAlive() and not t.isCritical()]

    @staticmethod
    def __do_shutdown__(_all_threads):

        from Ganga.Utility.logging import getLogger
        logger = getLogger('GangaThread')

        from Ganga.Core.GangaThread.WorkerThreads import _global_queues as queues

        if queues is not None:
            queues._purge_all()
            queues._stop_all_threads(shutdown=True)

            logger.debug("ExternalQueues still running: %s" % queues.threadStatus())

        logger.debug('Service threads to shutdown: %s' % ([i for i in reversed(list(_all_threads))]))

        logger.debug('Service threads to shutdown: %s' % ([i for i in reversed(list(_all_threads))]))

        # shutdown each individual threads in the pool
        nonCritThreads = []
        critThreads = []

        for t in _all_threads:
            if t.isCritical():
                critThreads.append(t)
            else:
                nonCritThreads.append(t)

        # while len( _all_threads ) != 0:
        # Shutdown NON critical threads first as these can cause some critical
        # threads to hang
        for t in reversed(nonCritThreads):
            logger.debug('shutting down Thread: %s' % t.getName())
            t.stop()
            logger.debug('shutdown Thread: %s' % t.getName())
            # t.unregister()

        # Shutdown critical threads now assuming that the non-critical ones
        # have disappeared
        for t in reversed(critThreads):
            logger.debug('shutting down Thread: %s' % t.getName())
            t.stop()
            logger.debug('shutdown Thread: %s' % t.getName())
            # t.unregister()

        #    nonCritThreads = []
        #    critThreads = []

        #    for t in _all_threads:
        #        if t.isCritical():
        #            critThreads.append( t )
        #        else:
        #            nonCritThreads.append( t )

        def __cnt_alive_threads__(_all_threads):
            num_alive_threads = 0
            for t in _all_threads:
                if t.isAlive():
                    num_alive_threads += 1
            return num_alive_threads

        num_alive_threads = __cnt_alive_threads__(_all_threads)

        while num_alive_threads > 0:
            from Ganga.Utility.logging import getLogger
            logger = getLogger('GangaThread')
            # fix for bug #62543 https://savannah.cern.ch/bugs/?62543
            # following 2 lines swapped so that we access no globals between
            # sleep and exit test
            num_alive_threads = __cnt_alive_threads__(_all_threads)
            logger.debug('number of alive threads: %d' % num_alive_threads)
            time.sleep(0.3)
            num_alive_threads = __cnt_alive_threads__(_all_threads)




"""Example action implementations.

See Demo for a simple example of how to reuse Core actions.

N.B. The examples do not form part of the public API and may be modified or
removed between releases.

"""


"""Ganga Robot is a tool for running a user-defined list of actions within the
context of a Ganga session.

The Framework package contains the driver code, see Framework.Driver, and the
action interface, see Framework.Action, which should be implemented by any
action to be executed by the driver.

The Lib package contains some implementations of the action interface, including
base abstract implementations, see Lib.Base, which provide infrastructure to
easily add custom submit/finish/extract/report actions, as well as some generic
submit/finish/extract/report actions, see Lib.Core, which are used for the
default configuration.

The scripts folder contains a robot script for launching the robot.

The exports folder contains files of exported jobs which can be specified for
submission via the CoreSubmitter_Patterns configuration option or as arguments
to the robot script.

For robot script usage:
 ganga --config-path=GangaRobot/ROBOT.INI robot help

"""

import Ganga.Utility.Config

def _initconfig():
    config=Ganga.Utility.Config.makeConfig('Robot','Parameters for the Robot to run repetitive tests',is_open=True)
    
    config.addOption('Driver_Run',['submit', 20, 'finish', 'extract', 'report'],
                     'List of action names and sleep periods (seconds)')
    config.addOption('Driver_Repeat',False,
                     'boolean indicating if the run should repeat indefinitely')
    
    config.addOption('Driver_Action_submit',
                     'GangaRobot.Lib.Core.CoreSubmitter.CoreSubmitter',
                     'GangaRobot.Framework.Action.IAction class names for submit action in Driver_Run')
    config.addOption('Driver_Action_finish',
                     'GangaRobot.Lib.Core.CoreFinisher.CoreFinisher',
                     'GangaRobot.Framework.Action.IAction class names for finish action in Driver_Run')
    config.addOption('Driver_Action_extract',
                     'GangaRobot.Lib.Core.CoreExtractor.CoreExtractor',
                     'GangaRobot.Framework.Action.IAction class names for extract action in Driver_Run')
    config.addOption('Driver_Action_report',
                     'GangaRobot.Lib.Core.CoreReporter.CoreReporter',
                     'GangaRobot.Framework.Action.IAction class names for report action in Driver_Run')

    config.addOption('BaseFinisher_Timeout',3600,
                     'Timeout (seconds) for waiting for jobs to finish')
    config.addOption('BaseExtractor_XmlFile',
                     '~/gangadir/robot/extract/${runid}.xml',
                     'Filename for XML extract data, ${runid} is replaced by current run id')
    config.addOption('BaseReporter_TextFile',
                     '~/gangadir/robot/report/${runid}.txt',
                     'Filename for TEXT report data, ${runid} is replaced by current run id')
    config.addOption('BaseReporter_HtmlFile',
                     '~/gangadir/robot/report/${runid}.html',
                     'Filename for HTML report data, ${runid} is replaced by current run id')

    config.addOption('CoreSubmitter_Patterns',
                     ['GangaRobot/exports/local-echo-jobs.txt'],
                     'Exported job file patterns. Can contain Unix-style glob pathname patterns (relative patterns are evaluated against the current working directory and the Ganga python root, i.e. ganga/python/)')
    config.addOption('CoreReporter_ExtractUrl','',
                     'URL for links to extract data (if empty no link is created)')
    
    config.addOption('FileEmailer_Host','localhost:25','SMTP host and port')
    config.addOption('FileEmailer_Type','html',
                     'Email type \'html\' (i.e. html + plain) or \'text\' (i.e. plain)')
    config.addOption('FileEmailer_From','',
                     'From address for email, e.g. sender@domain.org')
    config.addOption('FileEmailer_Recipients','',
                     'Recepient list for email, e.g. recipient1@domain.org, recipient2@domain.org (if empty no email is sent)')
    config.addOption('FileEmailer_Subject','Ganga Robot: ${runid}.',
                     'Subject for email, ${runid} is replaced by current run id')
    config.addOption('FileEmailer_TextFile','',
                     'Filename for TEXT email body, ${runid} is replaced by current run id')
    config.addOption('FileEmailer_HtmlFile','',
                     'Filename for HTML email body, ${runid} is replaced by current run id')
    config.addOption('ExceptionBehaviour','Fatal','Changes behaviour of robot when exception is thrown. Options are Continue, Break, and Fatal (Default)')

def loadPlugins(config={}):
    _initconfig()



import sys
import os

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

def assert_cannot_submit(j):
    from Ganga.GPIDev.Lib.Job.Job import JobError
    try:
        j.submit()
        assert False, 'submit() should raise JobError'
    except JobError:
        pass  

def assert_cannot_kill(j):
    from Ganga.GPIDev.Lib.Job.Job import JobError    
    try:
        j.kill()
        assert False, 'kill() should raise JobError'
    except JobError:
        pass  

def sleep_until_state(j, timeout=None, state='completed', break_states=None, sleep_period=1, verbose=False):
    '''
    Wait until the job reaches the specified state
    Returns:
     True: if the state has been reached in the given timeout period
     False: timeout occured or a break state has been reached
    If break_states is specified, the call terminates when job enters in one of these state, returning False
    '''

    from Ganga.GPIDev.Base.Proxy import stripProxy
    j = stripProxy(j)
    if j.master is not None:
        j = j.master

    if timeout is None:
        timeout = config['timeout']
        
    from time import sleep
    from Ganga.Core import monitoring_component
    from Ganga.Core.GangaRepository import getRegistryProxy
    
    jobs = getRegistryProxy('jobs')

    current_status = None
    while j.status != state and timeout > 0:
        if not monitoring_component.isEnabled():
            monitoring_component.runMonitoring(jobs=jobs.select(j.id,j.id))
        else:
            monitoring_component.alive = True
            monitoring_component.enabled = True
            monitoring_component.steps = -1
            monitoring_component.__updateTimeStamp = 0
            monitoring_component.__sleepCounter = -0.5
        if verbose and j.status != current_status:
            logger.info("Job %s: status = %s" % (str(j.id), str(j.status)))
        if current_status is None:
            current_status = j.status
        if type(break_states) == type([]) and j.status in break_states:
            logger.info("Job finished with status: %s" % j.status )
            return False
        sleep(sleep_period)
        timeout -= sleep_period
        logger.debug("Status: %s" % j.status)
    logger.info("Job finished with status: %s" % j.status )
    logger.info("Timeout: %s" % str(timeout))
    try:
        j._getRegistry().updateLocksNow()
    except:
        pass
    return j.status == state

def sleep_until_completed(j, timeout=None, sleep_period=1, verbose=False):
    return sleep_until_state(j, timeout, 'completed', ['new','killed','failed','unknown','removed'], verbose=verbose, sleep_period=sleep_period)

def is_job_state(j, states=None, break_states=None):
    if states is None:
        states = ['completed']
    #Allow the completed state to be a list of status.
    if j.status in states:
        return True
    else:
        if break_states:
            if type(break_states) == type([]):
                assert (j.status not in break_states), 'Job did not complete (Status = %s)' % j.status
                return False
        else:
            return False
            

def is_job_finished(j):
    ''' Once the job status has reached the final status, then True. '''
    return is_job_state(j, ['completed','new','killed','failed','unknown','removed'])

def is_job_completed(j):
    return is_job_state(j, ['completed'], ['new','killed','failed','unknown','removed']) 
        
def file_contains(filename,string):
    return file(filename).read().find(string) != -1

def write_file(filename,content):
    """ Open,write and close the file descriptor"""
    f = open(filename,'w')
    try: return f.write(content)
    finally: f.close()

def read_file(filename):
    """ read the file, and safely close the file at the end"""
    f = open(filename)
    try: return "\n%s\n" % f.read()
    finally: f.close()

import unittest
failureException = unittest.TestCase.failureException

try:
    from Ganga.Utility.Config import getConfig
    config = getConfig('TestingFramework')
except: # if we are outside Ganga, use a simple dict
    config={}




import os
import sys
import pickle
import traceback
from socket import *
import time
from commands import getstatusoutput

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

# Ganga Service class that provides the interface to the server
class GangaService:
    
    def __init__(self):
        self.port = -1 #43434
        self.timeout = 60 # in minutes
        self.gangadir = os.path.normpath(os.path.expandvars(os.path.expanduser("~/gangadir-server")))
        self.prerun = ""
        self.gangacmd = "ganga"
        self.userscript = None
        self.userscriptwaittime = 300
        pass

    def getServerInfo(self):
        """return a tuple of the hostname and port from the server.info file"""
        if os.path.exists( os.path.join(self.gangadir, "server", "server.info") ):
            host_port = open( os.path.join(self.gangadir, "server", "server.info") ).read().strip()
            if len(host_port.split(":")) == 1:
                hostname = host_port
                port = self.port
                if port == -1:
                    port = 43434
            else:
                hostname = host_port.split(":")[0]
                port = host_port.split(":")[1]

            return (hostname, port)
        return ("", -1)
        
    def killServer(self):
        """send a terminate request to the server"""
        
        # find out if server is local or remote
        if os.path.exists( os.path.join(self.gangadir, "server", "server.info") ):

            # watchdog file there so check timestamp
            host_port = self.getServerInfo()
            hostname = host_port[0]
            port = int(host_port[1])

            if (time.time() - os.path.getmtime(os.path.join(self.gangadir, "server", "server.info") )) < 60:
                if hostname != os.uname()[1]:
                    logger.info("Active server running on host '%s' - creating kill file" % hostname)
                    open(os.path.join(self.gangadir, "server", "server.kill"), "w").write("")
                    wait = 0
                    while os.path.exists(os.path.join(self.gangadir, "server", "server.info")) and wait < 30:
                        time.sleep(1)
                        wait += 1

                    if wait > 29:
                        logger.info("Could not kill server. Please kill manually on host '%s'" % hostname)
                        return False

                    # wait a little longer for the process to finish completely...
                    time.sleep(10)
                    
                    logger.info("Server kill signals sent though you should check the process has exited on host '%s'." % hostname)
                    return True
                else:
                    # try and connect to this port
                    addr = ('localhost',port)
                    sock = socket(AF_INET,SOCK_STREAM)
                    try:
                        sock.connect(addr)
                    except:
                        logger.info("Could not connect to server on port %d" % port)
                        return False

                    sock.send("###STOP###")
                    data = sock.recv(1024)
                    while data.find("###STOPPED###") == -1:
                        data += sock.recv(1024)
                    sock.close()

                    # check the process has gone away
                    logger.info("Stop signal sent. Waiting for Ganga process to finish...")
                    wait = 0
                    ret, out = getstatusoutput("ps -Af | grep ganga -i | grep %d" % port)
                    while ret == 0 and wait < 120:
                        time.sleep(1)
                        wait += 1
                        ret, out = getstatusoutput('ps -Af | grep ganga -i | grep -v "sh -c" | grep %d' % port)
                        
                    if wait > 119:
                        logger.info("Signals sent but ganga still running. Retry killServer or kill process manually.")
                        return False
                    
                    logger.info("Local server stopped")
                    return True
                                                                                                
            else:
                logger.info("Stale server file around. Removing...")
                os.system("rm %s" % os.path.join(self.gangadir, "server", "server.info"))
                return True
        else:
            logger.info("No Server running")
            return True

    def startServer(self):
        """Start the server if required"""

        # is there a server up? check the watchdog file
        if os.path.exists(os.path.join(self.gangadir, "server", "server.info")):

            host_port = self.getServerInfo()
            hostname = host_port[0]
            self.port = int(host_port[1])

            # watchdog file there so check timestamp
            if (time.time() - os.path.getmtime(os.path.join(self.gangadir, "server", "server.info"))) < 60:
                if hostname != os.uname()[1]:
                    logger.info("Active server running on host '%s'" % hostname)
                    return False
                else:
                    logger.info("Active server running on this machine.")
                    return True
            else:
                logger.info("Stale server file around. Removing...")
                os.system("rm %s" % os.path.join(self.gangadir, "server", "server.info"))

        # try to get an unsued port
        if self.port == -1:
            base_port = 40000
            import random
            random.seed()
            base_num = random.randint(0, 50) * 100
            
            # find any runnning ganga instances
            ret, out = getstatusoutput("ps -Af | grep ganga -i | wc -l")
            proc_offset = int(out) * 5

            # and a bit random for luck
            add_offset = random.randint(0, 4)
            
            # final port number
            self.port = base_port + base_num + proc_offset + add_offset
            logger.info("Trying port %d..." % self.port)

        # No, so start it
        logger.info("Starting server...")
        if self.userscript:
            cmd = "%s --daemon -o[PollThread]forced_shutdown_timeout=300 -o[Configuration]ServerPort=%d -o[Configuration]ServerTimeout=%d -o[Configuration]gangadir=%s -o[Configuration]ServerUserScript=%s -o[Configuration]ServerUserScriptWaitTime=%d %s/server-script.py" % (self.gangacmd, self.port, self.timeout, self.gangadir, 
                                                                                                                                                                                                                                                                                 self.userscript, self.userscriptwaittime,
                                                                                                                                                                                                                                                                                 os.path.join(os.path.dirname( os.path.abspath( __file__ ) ), "../Server" ) )
        else:
            cmd = "%s --daemon -o[PollThread]forced_shutdown_timeout=300 -o[Configuration]ServerPort=%d -o[Configuration]ServerTimeout=%d -o[Configuration]gangadir=%s %s/server-script.py" % (self.gangacmd, self.port, self.timeout, self.gangadir, 
                                                                                                                                                                                               os.path.join(os.path.dirname( os.path.abspath( __file__ ) ), "../Server" ) )
            
        if self.prerun != "":
            cmd = "%s && %s" % (self.prerun, cmd)

        logger.info("\nSettings: ")
        logger.info("  - port:               %d" % self.port)
        logger.info("  - timeout:            %d" % self.timeout)
        logger.info("  - gangadir:           %s" % self.gangadir)
        logger.info("  - prerun:             %s" % self.prerun)
        logger.info("  - gangacmd:           %s" % self.gangacmd)
        if self.userscript:
            logger.info("  - userscript:         %s" % self.userscript)
            logger.info("  - userscriptwaittime: %d" % self.userscriptwaittime)
        else:
            logger.info("  - default user script specified from .gangarc")
        
        logger.info("\nStarting server using command:\n%s" % cmd)
        os.system(cmd)
        
        # wait for the server file to be created
        logger.info("Waiting for server to initialise...")

        time_diff = 0
        while True:
            if os.path.exists(os.path.join(self.gangadir, "server", "server.info")):
                break

            time.sleep(1)
            time_diff += 1
            if time_diff > 30:
                logger.info("Server took too long to start.")
                return False
        
        return True

    def sendCmd(self, cmd):

        # check if server is up
        if not self.startServer():
            logger.info("Could not start the Ganga Server.")
            return ""
        
        # try and connect to this port
        addr = ('localhost',self.port)
        sock = socket(AF_INET,SOCK_STREAM)
        try:
            sock.connect(addr)
        except:
            logger.info("Could not connect to server on port %d" % self.port)
            return ""

        sock.send(cmd + "\n###ENDCMD###")
        logger.info("Command sent. Waiting for output from Ganga...")
        data = sock.recv(1024)
        while data.find("###ENDMSG###") == -1:
            data += sock.recv(1024)
        sock.close()

        return data.replace("###ENDMSG###", "")



from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
import time
from Ganga.GPIDev.Lib.Tasks import ITask
from GangaNA62.Lib.Tasks.NA62Transform import NA62Transform
from GangaNA62.Lib.Applications.NA62MC import NA62MC

########################################################################

class NA62Task(ITask):
    """NA62 add-ons for the Task framework"""
    _schema = Schema(Version(1,0), dict(ITask._schema.datadict.items() + {
        }.items()))
    
    _category = 'tasks'
    _name = 'NA62Task'
    _exportmethods = ITask._exportmethods + [ 'initFromString' ]

    _tasktype = "ITask"
    
    default_registry = "tasks"
    
    def initFromString(self, cfg_str, backend = None):
        """Initialize this Task with the config string. Format is:
        $prod_name|$chan|$radcor|$runs|$events|$mcversion|$script|$sites
        """
        
        if self.status != "new":
            logger.error("Cannot add more data to a new task yet. Give me time :)")
            return

        # extract and check params
        toks = cfg_str.split("|")
        if len(toks) != 9:
            logger.error("Error in format of input string %s" % cfg_str)
            return
        
        prod_name = toks[0]
        decay_chan = int(toks[1])
        if (toks[2] == "1"):
            radcor = True
        else:
            radcor = False
        num_runs = int(toks[3])
        num_events = int(toks[4])
        mc_version = int(toks[5])
        revision = int(toks[6])
        scr_name = toks[7]
        sites = [ s.strip() for s in toks[8].split(",") ]
            
        # create a transform
        trf = NA62Transform()
        trf.name = prod_name
        trf.num_jobs = num_runs

        if backend:
            trf.backend = backend._impl.clone()
            
        if (trf.backend._name == "LCG"):
            trf.backend.requirements.allowedSites = sites
            trf.backend.requirements.memory = 2000
        
        app = NA62MC()
        app.decay_type = decay_chan
        app.num_events = num_events
        app.decay_name = prod_name
        app.radcor = radcor
        app.mc_version = mc_version
        app.revision = revision
        app.script_name = scr_name
        trf.application = app

        self.appendTransform(trf)




# evaluate help expressions in the Ganga.GPI namespace
# because from ... import * may only be done at the module level
# hence we need this additional module to do the trick

from Ganga.GPI import *


def evaluate(v):
    try:
        return eval(v)
    except Exception as x:
        return v



#!/bin/bash

"exec" "python" "-u" "-Wignore" "$0" "$@"

"""
Run Athena

Usage:

$ source $SITEROOT/setup.sh
$ source $T_DISTREL/AtlasRelease/*/cmt/setup.sh -tag_add=???
$ runAthena -l [libraries] -r [rundir] -j [jobOs] -i [inputs] -o [optputs] -c \
            -p [pool_refs] -u [lrc_url]

-l [libraries] : an archive which contains libraries
-r [rundir]    : relative path to the directory where Athena runs
-j [jobOs]     : job options passed to athena. format: 'options'
-i [inputs]    : list of input files. format: ['in1',...'inN']
-o [outputs]   : map of output files. format: {'type':'name',..}
                  type:'hist','ntuple','ESD','AOD','TAG','AANT','Stream1','THIST'
-b             : bytestream
-c             : event collection
-p [pool_refs] : list of POOL refs
-u [lrc_url]   : URL of LRC
-f [fragment]  : jobO fragment
-a [jobO files]: archive name of jobOs
-m [minbias]   : list of minimum bias files
-n [cavern]    : list of cavern files
--debug        : debug
--directIn     : read input files from SE
--oldPrefix    : old prefix to be replaced when converting TURL
--newPrefix    : new prefix to be used when converting TURL

Example:

runAthena \
  -l libraries.tgz \
  -r PhysicsAnalysis/AnalysisCommon/UserAnalysis/UserAnalysis-00-03-03/run \
  -j "-c 'EvtMax=10' opt.py RecExCommon.py" \
  -i ['input1.AOD.pool.root','input2.AOD.pool.root','input3.AOD.pool.root'] \
  -o ['hist':'hist.root','ntuple':'ntuple.root','log':'athena.log']

Procedure:

* expand libraries
* make PoolFileCatalog.xml
* create post-jobO which overwrites some parameters
* get PDGTABLE.MeV
* run athena

"""

# import optparse
import re
import os
import sys
import time
import types
import getopt
import commands
import urllib
import pickle
import xml.dom.minidom
import threading

# error code
EC_PoolCatalog  = 20
EC_MissingArg   = 30
EC_AthenaFail   = 40
EC_NoInput      = 141
EC_MissingInput = 142
EC_ARA          = 143
EC_DBRelease    = 144
EC_Coll         = 145
EC_WGET         = 146
EC_LFC          = 147

# command-line parameters
eventColl  = False
byteStream = False
backNavi   = False
debugFlag  = False
poolRefs = []
urlLRC = ''
libraries    = ''
fragmentJobO = ''
archiveJobO  = ''
minbiasFiles = []
cavernFiles  = []
beamHaloFiles= []
beamGasFiles = []
oldPrefix    = ''
newPrefix    = ''
directIn     = False
lfcHost      = ''
inputGUIDs   = []
minbiasGUIDs = []
cavernGUIDs  = []
shipInput    = False
addPoolFC    = []
corCheck     = False
sourceURL    = 'https://gridui07.usatlas.bnl.gov:25443'
mcData       = ''
notSkipMissing = False
givenPFN     = False
runTrf       = False
envvarFile   = ''
runAra       = False
dbrFile      = ''
generalInput = False
guidBoundary = []
collRefName  = ''
useNextEvent = False
liveLog      = ''
dbrRun       = -1
useLocalIO   = False
codeTrace    = False
useFileStager= False
usePFCTurl   = False
copyTool     = ''
eventPickTxt = ''
eventPickSt  = 0
eventPickNum = -1
skipInputByRetry = []

opts, args = getopt.getopt(sys.argv[1:], "l:r:j:i:o:bcp:u:f:a:m:n:e",
                           ["pilotpars","debug","oldPrefix=","newPrefix=",
                            "directIn","lfcHost=","inputGUIDs=","minbiasGUIDs=",
                            "cavernGUIDs=","shipInput","addPoolFC=","corCheck",
                            "sourceURL=","mcData=","notSkipMissing","givenPFN",
                            "beamHalo=","beamGas=","trf","envvarFile=","ara",
                            "dbrFile=","generalInput","guidBoundary=",
                            "collRefName=","useNextEvent","liveLog=",
                            "dbrRun=","useLocalIO","codeTrace","useFileStager",
                            "usePFCTurl",
                            "accessmode=","copytool=",
                            "eventPickTxt=","eventPickSt=","eventPickNum=",
                            "skipInputByRetry=",
                            ])
for o, a in opts:
    if o == "-l":
        libraries=a
    if o == "-r":
        runDir=a
    if o == "-j":
        jobO=urllib.unquote(a)
    if o == "-i":
        exec("inputFiles="+a)
    if o == "-o":
        exec("outputFiles="+a)
    if o == "-m":
        exec("minbiasFiles="+a)
    if o == "-n":
        exec("cavernFiles="+a)
    if o == "--beamHalo":
        exec("beamHaloFiles="+a)
    if o == "--beamGas":
        exec("beamGasFiles="+a)
    if o == "-b":
        byteStream = True
    if o == "-c":
        eventColl = True
    if o == "-p":
        exec("poolRefs="+a)
    if o == "-u":
        urlLRC=a
    if o == "-f":
        fragmentJobO=a
    if o == "-a":
        archiveJobO=a
    if o == "-e":
        backNavi = True
    if o == "--debug":
        debugFlag = True
    if o == "--oldPrefix":
        oldPrefix = a
    if o == "--newPrefix":
        newPrefix = a
    if o == "--directIn":
        directIn = True
    if o == "--lfcHost":
        lfcHost = a
    if o == "--inputGUIDs":
        exec("inputGUIDs="+a)
    if o == "--minbiasGUIDs":
        exec("minbiasGUIDs="+a)
    if o == "--cavernGUIDs":
        exec("cavernGUIDs="+a)
    if o == "--shipInput":
        shipInput = True
    if o == "--addPoolFC":
        addPoolFC = a.split(',')
    if o == "--corCheck":
        corCheck = True
    if o == "--sourceURL":
        sourceURL = a
    if o == "--mcData":
        mcData = a
    if o == "--notSkipMissing":
        notSkipMissing = True
    if o == "--givenPFN":
        givenPFN = True
    if o == "--trf":
        runTrf = True
    if o == "--envvarFile":
        envvarFile = a
    if o == "--ara":
        runAra = True
    if o == "--dbrFile":
        dbrFile = a
    if o == "--generalInput":
        generalInput = True
    if o == "--guidBoundary":
        exec("guidBoundary="+a)
    if o == "--collRefName":
        collRefName = a
    if o == "--useNextEvent":
        useNextEvent = True
    if o == "--liveLog":
        liveLog = a
    if o == "--dbrRun":
        dbrRun = a
    if o == "--useLocalIO":
        useLocalIO = True
    if o == "--codeTrace":
        codeTrace = True
    if o == "--useFileStager":
        useFileStager = True
    if o == "--usePFCTurl":
        usePFCTurl = True
    if o == "--copytool":
        copyTool = a
    if o == "--eventPickTxt":
        eventPickTxt = a
    if o == "--eventPickSt":
        eventPickSt = int(a)
    if o == "--eventPickNum":
        eventPickNum = int(a)
    if o == "--skipInputByRetry":
        skipInputByRetry = a.split(',')
        

# save current dir
currentDir = os.getcwd()

# change full path
if envvarFile != '':
    envvarFile = '%s/%s' % (currentDir,envvarFile)

# dump parameter
try:
    print "=== parameters ==="
    print "libraries",libraries
    print "runDir",runDir
    print "jobO",jobO
    print "inputFiles",inputFiles
    print "outputFiles",outputFiles
    print "byteStream",byteStream
    print "eventColl",eventColl
    print "backNavi",backNavi
    print "debugFlag",debugFlag
    print "poolRefs",poolRefs
    print "urlLRC",urlLRC
    print "fragmentJobO",fragmentJobO
    print "minbiasFiles",minbiasFiles
    print "cavernFiles",cavernFiles
    print "beamHaloFiles",beamHaloFiles
    print "beamGasFiles",beamGasFiles
    print "oldPrefix",oldPrefix
    print "newPrefix",newPrefix
    print "directIn",directIn
    print "lfcHost",lfcHost
    print "inputGUIDs",inputGUIDs
    print "minbiasGUIDs",minbiasGUIDs
    print "cavernGUIDs",cavernGUIDs
    print "addPoolFC",addPoolFC
    print "corCheck",corCheck
    print "sourceURL",sourceURL
    print "mcData",mcData
    print "notSkipMissing",notSkipMissing
    print "givenPFN",givenPFN
    print "runTrf",runTrf
    print "envvarFile",envvarFile
    print "runAra",runAra
    print "dbrFile",dbrFile
    print "generalInput",generalInput
    print "liveLog",liveLog
    print "dbrRun",dbrRun
    print "useLocalIO",useLocalIO
    print "codeTrace",codeTrace
    print "useFileStager",useFileStager
    print "usePFCTurl",usePFCTurl
    print "copyTool",copyTool
    print "eventPickTxt",eventPickTxt
    print "eventPickSt",eventPickSt
    print "eventPickNum",eventPickNum
    print "skipInputByRetry",skipInputByRetry
    print "==================="    
except:
    sys.exit(EC_MissingArg)

# disable direct input for unsupported cases
if directIn:
    if useLocalIO:
        # use local IO
        directIn = False
        print "disabled directIn due to useLocalIO"        
    elif byteStream and newPrefix.startswith('root://'):
        # BS on xrootd
        directIn = False
        print "disabled directIn for xrootd/RAW"

# remove skipped files
if skipInputByRetry != []: 
    tmpInputList = []
    for tmpLFN in inputFiles:
        if not tmpLFN in skipInputByRetry:
            tmpInputList.append(tmpLFN)
    inputFiles = tmpInputList
    print "removed skipped files -> %s" % str(inputFiles)
        
# log watcher
class LogWatcher (threading.Thread):
    # onstructor
    def __init__(self,fileName,logName):
        threading.Thread.__init__(self)
        self.fileName = fileName
        self.logName  = logName
        self.offset   = 0
        self.lock     = threading.Lock()

    # terminate thread
    def terminate(self):
        self.lock.acquire()

    # update
    def update(self):
        try:
            import zlib
            import socket
            import httplib
            import mimetools
            # read log
            logFH = open(self.fileName)
            logFH.seek(self.offset)
            logStr = logFH.read()
            logFH.close()
            # upload
            if len(logStr) != 0:
                # compress
                zStr = zlib.compress(logStr)
                # construct HTTP request
                boundary = mimetools.choose_boundary()
                body  = '--%s\r\n' % boundary
                body += 'Content-Disposition: form-data; name="file"; filename="%s"\r\n' % self.logName
                body += 'Content-Type: application/octet-stream\r\n'
                body += '\r\n' + zStr + '\r\n'
                body += '--%s--\r\n\r\n' % boundary
                headers = {'Content-Type': 'multipart/form-data; boundary=%s' % boundary,
                           'Content-Length': str(len(body))}
                url = '%s/server/panda/updateLog' % sourceURL
                match = re.search('[^:/]+://([^/]+)(/.+)',url)
                host = match.group(1)
                path = match.group(2)
                # set timeout
                socket.setdefaulttimeout(60)
                # HTTPS connection
                conn = httplib.HTTPSConnection(host)
                conn.request('POST',path,body,headers)
                resp = conn.getresponse()
                data = resp.read()
                conn.close()
                print "updated LogWatcher at %s" % time.strftime('%Y-%m-%d %H:%M:%S',time.gmtime())
                # increment offset
                self.offset += len(logStr)                
        except:
            type, value, traceBack = sys.exc_info()
            print 'failed to update LogWatcher %s - %s' % (type,value)
        
    # main
    def run(self):
        print "start LogWatcher"
        while True:
            # update
            self.update()
            # check lock
            if self.lock.acquire(0):
                self.lock.release()
                time.sleep(60)
            else:
                # last update just in case
                time.sleep(10)
                self.update()
                # terminate
                print "terminate LogWatcher"
                return
            

# get PFNs from LRC
def _getPFNsFromLRC (urlLRC ,items,isGUID=True,old_prefix='',new_prefix=''):
    import urllib2
    # old prefix for regex
    old_prefix_re = old_prefix.replace('?','\?')
    pfnMap = {}
    if len(items)>0:
        # get PoolFileCatalog
        iITEM = 0
        strITEMs = ''
        for item in items:
            iITEM += 1
            # make argument
            strITEMs += '%s ' % item
            if iITEM % 35 == 0 or iITEM == len(items):
                # get PoolFileCatalog
                strITEMs = strITEMs.rstrip()
                if isGUID:
                    data = {'guids':strITEMs}
                else:
                    data = {'lfns':strITEMs}                    
                # avoid too long argument
                strITEMs = ''
                # GET
                url = '%s/lrc/PoolFileCatalog?%s' % (urlLRC,urllib.urlencode(data))
                req = urllib2.Request(url)
                fd = urllib2.urlopen(req)
                out = fd.read()
                if out.startswith('Error'):
                    continue
                if not out.startswith('<?xml'):
                    continue
                # get SURLs
                try:
                    root  = xml.dom.minidom.parseString(out)
                    files = root.getElementsByTagName('File')
                    for file in files:
                        # get ID
                        id = str(file.getAttribute('ID'))
                        # get PFN node
                        physical = file.getElementsByTagName('physical')[0]
                        pfnNode  = physical.getElementsByTagName('pfn')[0]
                        # convert UTF8 to Raw
                        pfn = str(pfnNode.getAttribute('name'))
                        # remove :8443/srm/managerv1?SFN=
                        pfn = re.sub(':8443/srm/managerv1\?SFN=','',pfn)
                        if old_prefix=='':
                            # remove protocol and host
                            pfn = re.sub('^[^:]+://[^/]+','',pfn)
                            # remove redundant /
                            pfn = re.sub('^//','/',pfn)
                            # put dcache if /pnfs
                            if pfn.startswith('/pnfs'):
                                pfn = 'dcache:%s' % pfn
                        else:
                            # check matching
                            if re.search(old_prefix_re,pfn) == None:
                                continue
                            # replace prefix
                            pfn = re.sub(old_prefix_re,new_prefix,pfn)
                        # append
                        pfnMap[id] = pfn
                except:
                    pass
    return pfnMap


# run LFC stuff in grid runtime
def _getPFNsFromLFC (lfcHost,guids, oldPrefix='', newPrefix='', qryMetadata=False, envvarFile=''):

    lfcCommand = """
import os
import re
import sys
import time
import pickle

# get PFNs from LFC
def _getPFNsFromLFC (lfc_host,items,old_prefix='',new_prefix='',getMeta=False):
    retVal = 0
    pfnMap = {}
    metaMap = {}
    # old prefix for regex
    old_prefix_re = old_prefix.replace('?','\?')
    # import lfc
    try:
        import lfc
    except:
        print "ERROR : cound not import lfc"
        retVal = 1
        return retVal,pfnMap
    # set LFC HOST
    os.environ['LFC_HOST'] = lfc_host
    # check bulk-operation
    if not hasattr(lfc,'lfc_getreplicas'):
        print "ERROR : bulk-ops is unsupported"
        retVal = 2
        return retVal,pfnMap
    frList = []
    # set nGUID for bulk-ops
    nGUID = 100
    iGUID = 0
    mapLFN = {}
    listGUID = []
    # loop over all items
    for item in items:
        iGUID += 1
        listGUID.append(item)
        if iGUID % nGUID == 0 or iGUID == len(items):
            # get replica
            nTry = 5
            for iTry in range(nTry):
                ret,resList = lfc.lfc_getreplicas(listGUID,'')
                if ret == 0 or iTry+1 == nTry:
                    break
                print "sleep due to LFC error"
                time.sleep(60)
            if ret != 0:
                err_num = lfc.cvar.serrno
                err_string = lfc.sstrerror(err_num)
                print "ERROR : LFC access failure - %s" % err_string
            else:
                for fr in resList:
                    if fr != None and ((not hasattr(fr,'errcode')) or \
                                       (hasattr(fr,'errcode') and fr.errcode == 0)):
                        print "replica found for %s" % fr.guid
                        # skip empty or corrupted SFN
                        if fr.sfn == '' or re.search('[^\w\./\-\+\?:&=]',fr.sfn) != None:
                            print "WARNING : wrong SFN '%s'" % fr.sfn
                            continue
                        print fr.sfn
                        # check host
                        if old_prefix != '' and (not fr.sfn.startswith(old_prefix)) and re.search(old_prefix_re,fr.sfn) == None:
                            continue
                        guid = fr.guid
                        # use first one
                        if pfnMap.has_key(guid):
                            onDiskFlag = False
                            for diskPath in ['/MCDISK/','/BNLT0D1/','/atlasmcdisk/','/atlasdatadisk/']:
                                if re.search(diskPath,fr.sfn) != None:
                                    onDiskFlag = True
                                    break
                            if not onDiskFlag:
                                continue
                            print "use disk replica"
                        if (old_prefix != '' and re.search(old_prefix_re,fr.sfn) == None) or old_prefix == '':
                            # conver to compact format
                            pfn = fr.sfn
                            pfn = re.sub('(:\d+)*/srm/v\d+/server\?SFN=','',pfn)
                            pfn = re.sub('(:\d+)*/srm/managerv\d+\?SFN=','',pfn)
                            # remove protocol and host
                            pfn = re.sub('[^:]+://[^/]+','',pfn)
                            pfn = new_prefix + pfn
                        else:
                            pfn = re.sub(old_prefix_re,new_prefix,fr.sfn)
                        # assign
                        pfnMap[guid] = pfn
                        # get metadata
                        if getMeta:
                            tmpMetaItem = {}
                            tmpMetaItem['size'] = fr.filesize
                            if fr.csumtype == 'AD':
                                tmpMetaItem['checksum'] = fr.csumvalue
                            # assign
                            metaMap[guid] = tmpMetaItem
            # reset
            listGUID = []
    # return PFN and metadata
    if getMeta:
        return retVal,{'pfn':pfnMap,'meta':metaMap}
    # return PFN only
    return retVal,pfnMap
"""

    directMetaMap = {}
    directTmp     = {}

    lfcPy = '%s/%s.py' % (os.getcwd(),commands.getoutput('uuidgen 2>/dev/null'))
    lfcOutPi = '%s/lfc.%s' % (os.getcwd(),commands.getoutput('uuidgen 2>/dev/null'))
    lfcPyFile = open(lfcPy,'w')
    lfcPyFile.write(lfcCommand)

    if qryMetadata:
        tmpGetPfnLfcStr = "st,out= _getPFNsFromLFC ('%s',%s,old_prefix='%s',new_prefix='%s',getMeta=True)\n"
    else:
        tmpGetPfnLfcStr = "st,out= _getPFNsFromLFC ('%s',%s,old_prefix='%s',new_prefix='%s')\n"

    lfcPyFile.write(tmpGetPfnLfcStr % (lfcHost,guids,oldPrefix,newPrefix))

    lfcPyFile.write("""
outPickFile = open('%s','w')
pickle.dump(out,outPickFile)
outPickFile.close()
sys.exit(st)
""" % lfcOutPi)

    lfcPyFile.close()

    # run LFC access in grid runtime
    lfcSh = '%s.sh' % commands.getoutput('uuidgen 2>/dev/null')

    if envvarFile != '':
        commands.getoutput('cat %s > %s' % (envvarFile,lfcSh))

    # check LFC module
    print "->check LFC.py"
    lfcS,lfcO = commands.getstatusoutput('python -c "import lfc"')
    print lfcS
    print lfcO
    if lfcS == 0:
        commands.getoutput('echo "python %s" >> %s' % (lfcPy,lfcSh))
    else:
        # use system python
        print "->use /usr/bin/python"
        commands.getoutput('echo "/usr/bin/python %s" >> %s' % (lfcPy,lfcSh))
    commands.getoutput('chmod +x %s' % lfcSh)
    tmpSt,tmpOut = commands.getstatusoutput('./%s' % lfcSh)
    print tmpSt
    print tmpOut
    # error check
    if re.search('ERROR : LFC access failure',tmpOut) != None:
        sys.exit(EC_LFC)

    if tmpSt == 0:
        lfcOutPiFile = open(lfcOutPi)
        directTmp = pickle.load(lfcOutPiFile)
        lfcOutPiFile.close()
        # decompose if pfn+meta is returned
        if 'meta' in directTmp:
            directMetaMap = directTmp['meta']
            directTmp = directTmp['pfn']
    else:
        directTmp     = {}
        directMetaMap = {}

    return directTmp, directMetaMap


pCopyToolName = 'pCopyToolWrapper'

pCopyToolStr = """#!/bin/bash

"exec" "python" "-u" "-Wignore" "$0" "$@"

import os
import sys
import time
import zlib
import commands

# change stdout
stdOutFD = os.open('pCopyToolWrapper.log',os.O_WRONLY|os.O_APPEND|os.O_CREAT)
os.dup2(stdOutFD, sys.stdout.fileno())

execfile('pCopyToolWrapperConf.py')

for arg in sys.argv[1:]:
    com += ' %s' % arg
lfilename = sys.argv[-1]
rfilename = sys.argv[-2]
nTry = 3    
for iTry in range(nTry):
    # copy
    print 'INFO: Try %s/%s : %s' % (iTry+1,nTry,com)
    fork_child_pid = 0
    fork_child_pid = os.fork()
    if fork_child_pid == -1:
        print "ERROR: failed to fork"
        sys.exit(1)
    if fork_child_pid == 0:
        # copy process
        status = os.system(com)
        status %= 255
        if status != 0:
            print "ERROR: failed to copy with" % status
            os._exit(status)
        # get adler32
        if not metaMap.has_key(rfilename):
            print "INFO: skip checksum verification"
            print "INFO: succeeded"
            print
            os._exit(0)
        radler32 = metaMap[rfilename]
        # calculate adler32
        sum1 = 1L
        try:
            f = open(lfilename, 'rb')
        except:
            print "ERROR: failed to open %s" % lfilename
            os._exit(1)
        for line in f:
            sum1 = zlib.adler32(line, sum1)
        f.close()
        # correction for bug 32 bit zlib
        if sum1 < 0:
            sum1 = sum1 + 2**32
        # convert to hex
        sum1 = "%08x" % sum1
        # compare
        if radler32 == sum1:
            print "INFO: checksum verified"
            print "INFO: succeeded"
            print                
            os._exit(0)
        # checksum failure
        print "ERROR: invalid checksum remote:%s local:%s" % (radler32,sum1)
        os._exit(1)
    else:
        # master process
        childExit    = False
        childStatus  = None
        timeoutValue = 900
        for iTimeOutTry in range(timeoutValue):
            # check child status
            pStat = os.waitpid(fork_child_pid,os.WNOHANG)
            if pStat[0] == 0:
                time.sleep(1)
            else:
                childExit = True
                childStatus = (pStat[1] & 0xFFFF) >> 8
                if childStatus == 0:
                    # return SUCCEEDED
                    sys.exit(0)
                break
        # kill child processes since it timed out
        if not childExit:
            print "ERROR: timeout"
            masterPID  = os.getpid()
            masterPGID = os.getpgrp()
            ppIDList = [fork_child_pid]
            psOutput = commands.getoutput('ps axjfww')
            for psLine in psOutput.split('\\n'):
                psItems = psLine.split()
                # get PPID
                try:
                    ppID = int(psItems[0])
                except:
                    ppID = None
                # get PID
                try:
                    pID = int(psItems[1])
                except:
                    pID = None
                # get PGID
                try:
                    pgID = int(psItems[2])
                except:
                    pgID = None
                # kill if it belongs to the child process
                if pgID == masterPGID and ppID in ppIDList:
                    # collect PPID
                    if not pID in ppIDList:
                        ppIDList.append(pID)
                    try:
                        # kill with SIGTERM
                        os.kill(pID,signal.SIGTERM)
                    except:
                        pass
                    try:
                        # kill with SIGKILL just in case
                        os.kill(pID,signal.SIGKILL)
                    except:
                        pass
            # wait for termination of child process
            try:
                os.wait()
            except:
                pass
        # cleanup and sleep
        commands.getoutput('rm -f %s' % lfilename)
        if iTry+1 < nTry:
            time.sleep(60)
# eventually failed
print "ERROR: failed to stage %s" % rfilename
print
sys.exit(1)
"""

#### H.C.LEE - wrapping up environment variables with double quote
def wrapEnvironVars( envvarFile='' ):
    # add "" in envvar
    try:
        newString = '#!/bin/bash\n'
        if envvarFile != '':
            tmpEnvFile = open(envvarFile)
            for line in tmpEnvFile:
                # remove \n
                line = line[:-1]
                match = re.search('([^=]+)=(.*)',line)
                if match != None:
                    # add ""
                    newString += '%s="%s"\n' % (match.group(1),match.group(2))
                else:
                    newString += '%s\n' % line
            tmpEnvFile.close()
            # overwrite
            tmpEnvFile = open(envvarFile,'w')
            tmpEnvFile.write(newString)
            tmpEnvFile.close()
    except:
        type, value, traceBack = sys.exc_info()
        print 'WARNING: changing envvar %s - %s' % (type,value)

wrapEnvironVars( envvarFile=envvarFile)


##### H.C. LEE - get files from PoolFileCatalog
def getFilesFromPFC():
    """
    parsing PFC.xml and put file guid, lfn, pfn into separate dictionaries:

    - guidMapFromPFC: using "lfn" as key, "guid" as value
    - directTmpTurl : using "guid" as key, "pfn" as value
    """
    guidMapFromPFC = {}
    directTmpTurl = {}
    try:
        print "===== PFC from pilot ====="
        tmpPcFile = open("PoolFileCatalog.xml")
        print tmpPcFile.read()
        tmpPcFile.close()
        # parse XML
        root  = xml.dom.minidom.parse("PoolFileCatalog.xml")
        files = root.getElementsByTagName('File')
        for file in files:
            # get ID
            id = str(file.getAttribute('ID'))
            # get PFN node
            physical = file.getElementsByTagName('physical')[0]
            pfnNode  = physical.getElementsByTagName('pfn')[0]
            # convert UTF8 to Raw
            pfn = str(pfnNode.getAttribute('name'))
            lfn = pfn.split('/')[-1]
            # append
            guidMapFromPFC[lfn] = id
            directTmpTurl[id] = pfn
    except:
        type, value, traceBack = sys.exc_info()
        print 'ERROR : Failed to collect GUIDs : %s - %s' % (type,value)

    return guidMapFromPFC, directTmpTurl

guidMapFromPFC, directTmpTurl = getFilesFromPFC()
print "===== GUIDs in PFC ====="
print guidMapFromPFC

### H.C.LEE - resolve inputfile information
def resolveInputFiles(directTmpTurl=directTmpTurl):
    """
     - 3 cases: "directIn", "givenPFN", or others

     - for "directIn":
       * usePFCTurl (LFN:PFN given directly in PFC.xml)
       * getting PFN from LFC if lfcHost is given (in FS case, file metadata also queried from LFC)
       * getting PFN from LRC otherwise

     - for "givenPFN":
       * aggregate file LFN given by user

     - for "others":
       * get file LFN by listing current directory
    """
    directPfnMap = {}
    directMetaMap = {}
    if directIn:
        if usePFCTurl:
            # Use the TURLs from PoolFileCatalog.xml created by pilot
            print "===== GUIDs and TURLs in PFC ====="
            print directTmpTurl
            directTmp = directTmpTurl
        else:
            if lfcHost != '':

                directTmp, directMetaMap = _getPFNsFromLFC(lfcHost=lfcHost,
                                                           guids=inputGUIDs+minbiasGUIDs+cavernGUIDs,
                                                           oldPrefix=oldPrefix,
                                                           newPrefix=newPrefix,
                                                           qryMetadata=useFileStager,
                                                           envvarFile=envvarFile)
            else:
                # get PFNs from LRC
                directTmp = _getPFNsFromLRC (urlLRC,inputFiles+minbiasFiles+cavernFiles,
                                             isGUID=False,old_prefix=oldPrefix,
                                             new_prefix=newPrefix)
        # collect LFNs
        curFiles   = []
        directPFNs = {}
        for id in directTmp.keys():
            lfn = directTmp[id].split('/')[-1]
            curFiles.append(lfn)
            directPFNs[lfn] = directTmp[id]
        directPfnMap = directTmp
    elif givenPFN:
        # collect LFNs
        curFiles   = []
        for lfn in inputFiles+minbiasFiles+cavernFiles+beamHaloFiles+beamGasFiles:
            curFiles.append(lfn)
    else:
        curFiles = os.listdir('.')
        
    return directPfnMap,directMetaMap,directPFNs,curFiles

directPfnMap,directMetaMap,directPFNs,curFiles = resolveInputFiles(directTmpTurl=directTmpTurl)

#### H.C.LEE - update the input files with exactly the PFNs
def updateInputFiles( inputFiles, curFiles, directPFNs ):
    """
    update the given inputFiles with PFN given the previously resolved input file information
    """
    tmpFiles = tuple(inputFiles)

    for tmpF in tmpFiles:
        findF = False
        findName = ''
        for curF in curFiles:
            if re.search('^'+tmpF,curF) != None:
                findF = True
                findName = curF
                break
        # remove if not exist
        if not findF:
            print "%s not exist" % tmpF
            inputFiles.remove(tmpF)
        # use URL
        if directIn and findF:
            inputFiles.remove(tmpF)
            inputFiles.append(directPFNs[findName])
    if len(inputFiles) == 0:
        print "No input file is available"
        sys.exit(EC_NoInput)
    if notSkipMissing and len(inputFiles) != len(tmpFiles):
        print "Some input files are missing"
        sys.exit(EC_MissingInput)


flagMinBias = False
flagCavern  = False
flagBeamGas = False
flagBeamHalo= False

if len(inputFiles) > 0 and (not shipInput):
    updateInputFiles( inputFiles=inputFiles, curFiles=curFiles, directPFNs=directPFNs )
                         
if len(minbiasFiles) > 0:
    flagMinBias = True
    updateInputFiles( inputFiles=minbiasFiles, curFiles=curFiles, directPFNs=directPFNs )
        
if len(cavernFiles) > 0:
    flagCavern = True
    updateInputFiles( inputFiles=cavernFiles, curFiles=curFiles, directPFNs=directPFNs )

if len(beamHaloFiles) > 0:
    flagBeamHalo = True
    updateInputFiles( inputFiles=beamHaloFiles, curFiles=curFiles, directPFNs=directPFNs )

if len(beamGasFiles) > 0:
    flagBeamGas = True
    updateInputFiles( inputFiles=beamGasFiles, curFiles=curFiles, directPFNs=directPFNs )

print "=== New inputFiles ==="
print inputFiles
if flagMinBias:
    print "=== New minbiasFiles ==="
    print minbiasFiles
if flagCavern:    
    print "=== New cavernFiles ==="
    print cavernFiles
if flagBeamHalo:
    print "=== New beamHaloFiles ==="
    print beamHaloFiles
if flagBeamGas:
    print "=== New beamGasFiles ==="
    print beamGasFiles

#### H.C.LEE - preparing the workspace for running athena user code
def prepareWorkspace():
    """
    - create workdir
    - expand libraries
    - expend joboption archive
    - create cmtdir
    - resolve Athena version

    returns workDir,cmtDir,athenaVer
    """

    workDir = currentDir+"/workDir"
    commands.getoutput('rm -rf %s' % workDir)
    os.makedirs(workDir)
    os.chdir(workDir)

    # expand libraries
    if libraries == '':
        pass
    elif libraries.startswith('/'):
        print commands.getoutput('tar xvfzm %s' % libraries)
    else:
        print commands.getoutput('tar xvfzm %s/%s' % (currentDir,libraries))

    # expand jobOs if needed
    if archiveJobO != "":
        print "--- wget for jobO ---"
        output = commands.getoutput('wget -h')
        wgetCommand = 'wget'
        for line in output.split('\n'):
            if re.search('--no-check-certificate',line) != None:
                wgetCommand = 'wget --no-check-certificate'
                break
        com = '%s %s/cache/%s' % (wgetCommand,sourceURL,archiveJobO)
        nTry = 3
        for iTry in range(nTry):
            print 'Try : %s' % iTry
            status,output = commands.getstatusoutput(com)
            print output
            if status == 0:
                break
            if iTry+1 == nTry:
                print "ERROR : cound not get jobO files from panda server"
                sys.exit(EC_WGET)
            time.sleep(30)
        print commands.getoutput('tar xvfzm %s' % archiveJobO)

    # make rundir just in case
    commands.getoutput('mkdir %s' % runDir)
    # go to run dir
    os.chdir(runDir)

    # make cmt dir
    cmtDir = '%s/%s/cmt' % (workDir,commands.getoutput('uuidgen 2>/dev/null'))
    commands.getoutput('mkdir -p %s' % cmtDir)

    # get athena version
    com  = 'cd %s;' % cmtDir
    com += 'echo "use AtlasPolicy AtlasPolicy-*" > requirements;'
    com += 'cmt config;'
    com += 'source ./setup.sh;'
    com += 'cd -;'
    com += 'cmt show projects'
    out = commands.getoutput(com)
    athenaVer = None
    try:
        for line in out.split('\n'):
            if not line.startswith('#'):
                items = line.split()
                if items[0] in ('dist','AtlasRelease','AtlasOffline'):
                    athenaVer = items[1]
                    break
    except:
        print out
        print "ERROR : cannot get Athena Version"

    print "Athena Ver:%s" % athenaVer

    return workDir,cmtDir,athenaVer

workDir,cmtDir,athenaVer = prepareWorkspace()

# create PoolFC
def _createPoolFC(pfnMap):
    outFile = open('PoolFileCatalog.xml','w')
    # write header
    header = \
    """<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
    <!-- Edited By POOL -->
    <!DOCTYPE POOLFILECATALOG SYSTEM "InMemory">
    <POOLFILECATALOG>
    """
    outFile.write(header)
    # write files
    item = \
    """
      <File ID="%s">
        <physical>
          <pfn filetype="ROOT_All" name="%s"/>
        </physical>
        <logical/>
      </File>
    """
    for guid,pfn in pfnMap.iteritems():
        outFile.write(item % (guid.upper(),pfn))
    # write trailer
    trailer = \
    """
    </POOLFILECATALOG>
    """
    outFile.write(trailer)
    outFile.close()
    

# build pool catalog

print "build pool catalog"
commands.getoutput('rm -f PoolFileCatalog.xml')
# for input files
if eventColl and mcData == '':
    # ROOT ver collection or AANT
    if len(inputFiles)>0:
        # use Collection Tools
        if guidBoundary != [] and collRefName != '':
            newInputFiles = []
            for fileName in inputFiles:
                # split
                com = 'CollSplitByGUID.exe -splitref %s -src PFN:%s RootCollection' % \
                      (collRefName,fileName)
                print com
                status,out = commands.getstatusoutput(com)
                if status != 0:
                    print out
                    print status
                    print "Failed to run %s" % com
                    sys.exit(EC_Coll)
                # get sub_collection_*
                subCollectionMap = {}
                filesInCurDir = os.listdir('.') 
                for tmpName in filesInCurDir:
                    match = re.search('sub_collection_(\d+)\.root',tmpName)
                    if  match != None:
                        subCollectionMap[int(match.group(1))] = tmpName
                # sort
                subCollection = subCollectionMap.keys()
                subCollection.sort()
                for idxSubColl in subCollection:
                    tmpName = subCollectionMap[idxSubColl]
                    # check if corresponding GUID is there
                    com = 'CollListFileGUID.exe -queryopt %s -src PFN:%s RootCollection' % \
                          (collRefName,tmpName)
                    print com
                    status,out = commands.getstatusoutput(com)
                    print out
                    if status != 0:
                        print status
                        print "Failed to run %s" % com
                        sys.exit(EC_Coll)
                    # look for GUID
                    foundGUID = False
                    for tmpGUID in guidBoundary:
                        if out.find(tmpGUID) != -1:
                            foundGUID = True
                            break
                    if foundGUID:
                        # rename
                        newName = '%s.%s' % (fileName,tmpName)
                        os.rename(tmpName,newName)
                        # append
                        newInputFiles.append(newName)
                    else:
                        # remove
                        os.remove(tmpName)
            # use new list
            if newInputFiles == []:
                print "Empty input list after CollXYZFileGUID.exe"
                sys.exit(EC_Coll)
            inputFiles = newInputFiles
            # add references
            poolRefs += guidBoundary
        else:
            # get extPoolRefs.C
            macroPoolRefs = 'extPoolRefs.C'
            # append workdir to CMTPATH
            env = 'CMTPATH=%s:$CMTPATH' % workDir
            com  = 'export %s;' % env
            com += 'cd %s;' % cmtDir
            com += 'echo "use AtlasPolicy AtlasPolicy-*" > requirements;'
            com += 'cmt config;'
            com += 'source ./setup.sh;'
            com += 'cd -;'
            com += 'get_files -jo %s' % macroPoolRefs
            print commands.getoutput(com)
            for fileName in inputFiles:
                # build ROOT command
                com = 'echo '
                if not directIn:
                    # form symlink to input file without attemptNr
                    newFileName = re.sub('\.\d+$','',fileName)
                    try:
                        os.symlink('%s/%s' % (currentDir,fileName),newFileName)
                    except:
                        pass
                    # create symlink with attemptNr for new Athena version
                    try:
                        os.symlink('%s/%s' % (currentDir,fileName),fileName)
                    except:
                        pass
                    # just in case for shipInput
                    if shipInput:
                        try:
                            os.rename(fileName,newFileName)
                        except:
                            pass
                        try:
                            os.symlink(newFileName,fileName)
                        except:
                            pass
                else:
                    # direct reading from SE
                    newFileName = fileName
                    
                if athenaVer == None or athenaVer < "14.4.0":
                    # use old macro
                    com += '%s ' % newFileName
                    com += ' --- | root.exe -b %s' % macroPoolRefs
                    print com
                    status,output = commands.getstatusoutput(com)
                    print output
                    # get POOL refs
                    for line in output.split('\n'):
                        if line.startswith('PoolRef:') or line.startswith('ESD Ref:') or \
                               line.startswith('RDO Ref:') or line.startswith('St1 Ref:') or \
                               line.startswith('RAW Ref:'):
                            match = re.search('\[DB=([^\]]+)\]',line)
                            if match != None:
                                if not match.group(1) in poolRefs:
                                    poolRefs.append(match.group(1))
                else:
                    # use CollListFileGUID for 14.4.0 onward
                    com = 'CollListFileGUID.exe -counts -src PFN:%s RootCollection' % newFileName
                    print com
                    status,output = commands.getstatusoutput(com)
                    print output
                    # get POOL refs
                    for line in output.split('\n'):
                        tmpItems = line.split()
                        if len(tmpItems) != 3:
                            continue
                        if re.search('^\w{8}-\w{4}-\w{4}-\w{4}-\w{12}$',tmpItems[1]) != None:
                            if not tmpItems[1] in poolRefs:
                                poolRefs.append(tmpItems[1])


    # new poolRefs
    print "=== New poolRefs ==="
    print poolRefs
    if len(poolRefs)>0:
        if lfcHost != '':
            # get PFNs from LFC
            pfnMap,dummy = _getPFNsFromLFC(lfcHost=lfcHost,
                                           guids=poolRefs,
                                           oldPrefix=oldPrefix,
                                           newPrefix=newPrefix,
                                           qryMetadata=False,
                                           envvarFile=envvarFile)
        else:
            # get PFNs from LRC 
            pfnMap = _getPFNsFromLRC (urlLRC,poolRefs,isGUID=True,
                                      old_prefix=oldPrefix,
                                      new_prefix=newPrefix)

        print "=== Create PoolFC ==="
        for ref in poolRefs:
            if ref not in pfnMap:
                print " %s not found" % ref
                
        # create PoolFC
        _createPoolFC(pfnMap)

elif len(inputFiles+minbiasFiles+cavernFiles+beamHaloFiles+beamGasFiles) > 0:
    # POOL or BS files
    filesToPfcMap = {}
    for fileName in inputFiles+minbiasFiles+cavernFiles+beamHaloFiles+beamGasFiles:
        if (not directIn) and (not givenPFN):
            targetName = fileName
            # for rome data
            if re.search(fileName,'\.\d+$')==None and (not fileName in curFiles):
                for cFile in curFiles:
                    if re.search('^'+fileName,cFile) != None:
                        targetName = cFile
                        break
            # form symlink to input file
            try:
                os.symlink('%s/%s' % (currentDir,targetName),fileName)
            except:
                pass
        if (not byteStream) and mcData == '' and (not generalInput) and not (runTrf and not runAra):
            # corruption check by scanning all TTrees
            if corCheck:
                # construct command
                print "=== check corruption for %s ===" % fileName
                optPy = '%s.py' % commands.getoutput('uuidgen 2>/dev/null')
                outFile = open(optPy,'w')
                outFile.write("""
import sys
import ROOT
try:
    import AthenaROOTAccess.transientTree
except:
    pass
t = ROOT.TFile('%s')
""" % fileName)
                outFile.write("""
keyList = t.GetListOfKeys()
scannedKeys = []
for keyItem in keyList:
    tree = keyItem.ReadObj()
    if tree.GetName() in scannedKeys:
        continue
    print '===%s===' % tree.GetName()
    if tree.GetName() == 'CollectionTree':
        try:
            import AthenaROOTAccess.transientTree
            tree = ROOT.AthenaROOTAccess.TChainROOTAccess('CollectionTree')
        except:
            pass
    nEvent = tree.GetEntriesFast()
    print nEvent
    detectFlag = False
    try:
        for iEvent in range(nEvent):
            ret = tree.GetEntry(iEvent)
            if ret < 0:
                print ret
                print iEvent
                detectFlag = True
                sys.exit(1)
    except:
        if detectFlag:
            sys.exit(1)
        type, value, traceBack = sys.exc_info()
        print 'EXCEPT: %s - %s' % (type,value)
        if 'St9bad_alloc' in str(value):
            sys.exit(2)
    scannedKeys.append(tree.GetName())
sys.exit(0)
""")
                outFile.close()
                # run checker
                status,out = commands.getstatusoutput('python %s' % optPy)
                print status
                print out
                commands.getoutput('rm -f %s' % optPy)
                if status != 0 \
                       or out.find("read too few bytes") != -1 \
                       or out.find("read too many bytes") != -1 \
                       or out.find("segmentation violation") != -1:
                    print "->skip %s" % fileName
                    continue
            # insert it to pool catalog
            tmpLFNforPFC = fileName.split('/')[-1]
            tmpLFNforPFC = re.sub('__DQ2-\d+$','',tmpLFNforPFC)
            if tmpLFNforPFC in guidMapFromPFC:
                filesToPfcMap[guidMapFromPFC[tmpLFNforPFC]] = fileName
            elif not givenPFN:
                print "ERROR : %s not found in the pilot PFC" % fileName
        # create PFC for directIn + trf
        if directIn and (runTrf and not runAra):
            _createPoolFC(directPfnMap)
            # form symlink to input file mainly for DBRelease
            for tmpID in directPfnMap.keys():
                lfn = directPfnMap[tmpID].split('/')[-1] 
                try:
                    targetName = '%s/%s' % (currentDir,lfn)
                    if os.path.exists(targetName):
                        os.symlink(targetName,lfn)
                except:
                    pass
    # create PFC for local files
    if filesToPfcMap != {}: 
        _createPoolFC(filesToPfcMap)
    elif givenPFN:
        # insert using pool_insertFTC since GUIDs are unavailabe from the pilot 
        for fileName in inputFiles+minbiasFiles+cavernFiles+beamHaloFiles+beamGasFiles:
            com = 'pool_insertFileToCatalog %s' % fileName
            print com
            os.system(com)
        
    # read PoolFileCatalog.xml
    pLines = ''
    try:
        pFile = open('PoolFileCatalog.xml')
        for line in pFile:
            pLines += line
        pFile.close()
    except:
        if mcData == '' and not (runTrf and not runAra):
            print "ERROR : cannot open PoolFileCatalog.xml"
    # remove corrupted files
    print "=== corruption check ==="
    # doesn't check BS/nonRoot files since they don't invoke insert_PFC
    if (not byteStream) and mcData == '' and (not generalInput) and not (runTrf and not runAra):
        tmpFiles = tuple(inputFiles)
        for tmpF in tmpFiles:
            if re.search(tmpF,pLines) == None:
                inputFiles.remove(tmpF)
                print "%s is corrupted or non-ROOT file" % tmpF
        if notSkipMissing and len(inputFiles) != len(tmpFiles):
            print "Some input files are missing"
            sys.exit(EC_MissingInput)
    if len(inputFiles)==0:        
        print "No input file is available after corruption check"
        sys.exit(EC_NoInput)        
    # extract POOL refs
    if backNavi:
        # construct command
        evtPy = 'EventCount.py'
        optPy = '%s.py' % commands.getoutput('uuidgen 2>/dev/null')
        print "=== run %s to extract POOL refs ===" % evtPy
        com  = 'get_files -jo %s;' % evtPy
        com += 'echo "theApp.EvtMax=-1" > %s;' % optPy
        com += 'athena.py -c "In=%s" %s %s;' % (inputFiles,evtPy,optPy)
        com += 'rm -f %s' % optPy
        # run athena
        status,out = commands.getstatusoutput(com)
        print out
        # extract
        flagStream = False
        for line in out.split('\n'):
            if re.search('^EventCount',line) == None:
                continue
            if re.search("Input contained references to the following File GUID",line) != None:
                flagStream = True
                continue
            if re.search("Input contained the following CLIDs and Keys",line) != None:
                flagStream = False
                continue
            if flagStream:
                match = re.search('([0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})',
                                  line)
                if match != None:
                    poolRefs.append(match.group(1))


        # new poolRefs
        print "=== New poolRefs ==="
        print poolRefs
        if len(poolRefs)>0:
            if lfcHost != '':
                # get PFNs from LFC
                pfnMap,dummy = _getPFNsFromLFC(lfcHost=lfcHost,
                                               guids=poolRefs,
                                               oldPrefix=oldPrefix,
                                               newPrefix=newPrefix,
                                               qryMetadata=False,
                                               envvarFile=envvarFile)
            else:
                # get PFNs from LRC 
                pfnMap = _getPFNsFromLRC (urlLRC,poolRefs,isGUID=True,
                                          old_prefix=oldPrefix,
                                          new_prefix=newPrefix)

            print "=== add POOL refs to PoolFC ==="
            for ref in poolRefs:
                if ref not in pfnMap:
                    print " %s not found" % ref

            # extract FIDs from PoolFC
            try:
                root  = xml.dom.minidom.parse("PoolFileCatalog.xml")
                files = root.getElementsByTagName('File')
                for file in files:
                    # get ID
                    id = str(file.getAttribute('ID'))
                    # get PFN node
                    physical = file.getElementsByTagName('physical')[0]
                    pfnNode  = physical.getElementsByTagName('pfn')[0]
                    # convert UTF8 to Raw
                    pfn = str(pfnNode.getAttribute('name'))
                    # append
                    pfnMap[id] = pfn
            except:
                pass
            
            # create PoolFC
            _createPoolFC(pfnMap)

# for user specified files
for fileName in addPoolFC:
    # insert it to pool catalog    
    com = 'pool_insertFileToCatalog %s' % fileName
    print com
    status,output = commands.getstatusoutput(com)
    print output

# print PoolFC
print "=== PoolFileCatalog.xml ==="
print commands.getoutput('cat PoolFileCatalog.xml')
print

# create symlink for MC
if mcData != '' and len(inputFiles) != 0:
    print "=== make symlink for %s ===" % mcData
    # expand mcdata.tgz
    commands.getoutput('rm -f %s' % mcData)
    status,output = commands.getstatusoutput('tar xvfzm %s' % inputFiles[0])
    print output
    if status != 0:
        print "ERROR : MC data corrupted"
        sys.exit(EC_NoInput)
    # look for .dat
    foundMcData = False
    for line in output.split('\n'):
        if line.endswith('.dat'):
            status,output = commands.getstatusoutput('ln -fs %s %s' % \
                                                     (line.split()[-1],mcData))
            if status != 0:
                print output                
                print "ERROR : failed to create symlink for MC data"
                sys.exit(EC_NoInput)
            foundMcData = True
            break
    if not foundMcData:
        print "ERROR : cannot find *.dat in %s" % inputFiles[0]
        sys.exit(EC_NoInput)

# setup DB/CDRelease
if dbrFile != '':
    if dbrRun == -1:
        print "=== setup DB/CDRelease (old style) ==="
        # expand 
        status,out = commands.getstatusoutput('tar xvfzm %s/%s' % (currentDir,dbrFile))
        print out
        # remove
        print commands.getstatusoutput('rm %s/%s' % (currentDir,dbrFile))
    else:
        print "=== setup DB/CDRelease (new style) ==="
        # make symlink
        print commands.getstatusoutput('ln -fs %s/%s %s' % (currentDir,dbrFile,dbrFile))
        # run Reco_trf and set env vars
        dbCom = 'Reco_trf.py RunNumber=%s DBRelease=%s' % (dbrRun,dbrFile)
        print dbCom
        status,out = commands.getstatusoutput(dbCom)
        print out
        # remove
        print commands.getstatusoutput('rm %s/%s' % (currentDir,dbrFile))
    # look for setup.py
    tmpSetupDir = None
    for line in out.split('\n'):
        if line.endswith('setup.py'):
            tmpSetupDir = re.sub('setup.py$','',line)
            break
    # check
    if tmpSetupDir == None:
        print "ERROR : cound not find setup.py in %s" % dbrFile
        sys.exit(EC_DBRelease)
    # run setup.py
    dbrSetupStr  = "import os\nos.chdir('%s')\nexecfile('setup.py',{})\nos.chdir('%s')\n" % \
                   (tmpSetupDir,os.getcwd())
    dbrSetupStr += "import sys\nsys.stdout.flush()\nsys.stderr.flush()\n"
        
               
# create post-jobO file which overwrites some parameters
postOpt = 'post_' + commands.getoutput('uuidgen 2>/dev/null') + '.py'
oFile = open(postOpt,'w')
oFile.write("""
try:
    EventSelectorAthenaPool.__getattribute__ = orig_ESAP__getattribute
except:
    pass

def _Service(str):
    try:
        svcMgr = theApp.serviceMgr()
        return getattr(svcMgr,str)
    except:
        return Service(str)
""")
if len(inputFiles) != 0 and mcData == '' and not runAra:
    if (re.search('theApp.EvtMax',fragmentJobO) == None) and \
       (re.search('EvtMax',jobO) == None):
        oFile.write('theApp.EvtMax = -1\n')
    if byteStream:
        # BS
        oFile.write('ByteStreamInputSvc = _Service( "ByteStreamInputSvc" )\n')
        oFile.write('ByteStreamInputSvc.FullFileName = %s\n' % inputFiles)        
    else:
        oFile.write('EventSelector = _Service( "EventSelector" )\n')
        if eventColl:
            # TAG
            newInputs = []
            for infile in inputFiles:
                # remove suffix for event collection
                if athenaVer == None or athenaVer < "13.0.30":
                    newInputs.append(re.sub('\.root\.*\d*$','',infile))
                elif (not directIn):
                    # symlinks don't have attemptNr
                    newInputs.append(re.sub('\.\d+$','',infile))                
                else:
                    # directIn doen't create symlinks
                    newInputs.append(infile)
            oFile.write('EventSelector.InputCollections = %s\n' % newInputs)
            oFile.write('EventSelector.CollectionType = "ExplicitROOT"\n')
        else:
            # normal POOL
            oFile.write('EventSelector.InputCollections = %s\n' % inputFiles)
elif len(inputFiles) != 0 and runAra:
    for tmpInput in inputFiles:
        oFile.write("CollectionTree.Add('%s')\n" % tmpInput)
if flagMinBias:
    oFile.write('minBiasEventSelector = _Service( "minBiasEventSelector" )\n')
    oFile.write('minBiasEventSelector.InputCollections = %s\n' % minbiasFiles)
if flagCavern:
    oFile.write('cavernEventSelector = _Service( "cavernEventSelector" )\n')
    oFile.write('cavernEventSelector.InputCollections = %s\n' % cavernFiles)
if flagBeamHalo:
    oFile.write('BeamHaloEventSelector = _Service( "BeamHaloEventSelector" )\n')
    oFile.write('BeamHaloEventSelector.InputCollections = %s\n' % beamHaloFiles)
if flagBeamGas:
    oFile.write('BeamGasEventSelector = _Service( "BeamGasEventSelector" )\n')
    oFile.write('BeamGasEventSelector.InputCollections = %s\n' % beamGasFiles)
if 'hist' in outputFiles:    
    oFile.write('HistogramPersistencySvc=_Service("HistogramPersistencySvc")\n')
    oFile.write('HistogramPersistencySvc.OutputFile = "%s"\n' % outputFiles['hist'])
if 'ntuple' in outputFiles:
    oFile.write('NTupleSvc = _Service( "NTupleSvc" )\n')
    firstFlag = True
    for sName,fName in outputFiles['ntuple']:
        if firstFlag:
            firstFlag = False
            oFile.write('NTupleSvc.Output=["%s DATAFILE=\'%s\' OPT=\'NEW\'"]\n' % (sName,fName))            
        else:
            oFile.write('NTupleSvc.Output+=["%s DATAFILE=\'%s\' OPT=\'NEW\'"]\n' % (sName,fName))
oFile.write("""
_configs = []
seqList = []
pTmpStreamList = []
try:
    from AthenaCommon.AlgSequence import AlgSequence
    tmpKeys = AlgSequence().allConfigurables.keys()
    # get AlgSequences
    seqList = [AlgSequence()]
    try:
        for key in tmpKeys:
            # check if it is available via AlgSequence
            if not hasattr(AlgSequence(),key.split('/')[-1]):
                continue
            # get full name
            tmpConf = getattr(AlgSequence(),key.split('/')[-1])
            if hasattr(tmpConf,'getFullName'):
                tmpFullName = tmpConf.getFullName()
                # append AthSequencer
                if tmpFullName.startswith('AthSequencer/'):
                    seqList.append(tmpConf)
    except:
        pass
    # loop over all sequences
    for tmpAlgSequence in seqList:
        for key in tmpKeys:
            if key.find('/') != -1:
                key = key.split('/')[-1]
            if hasattr(tmpAlgSequence,key):    
                _configs.append(key)
except:
    pass

def _getConfig(key):
    if seqList == []:
        from AthenaCommon.AlgSequence import AlgSequence
        return getattr(AlgSequence(),key)
    else:
        for tmpAlgSequence in seqList:
            if hasattr(tmpAlgSequence,key):
                return getattr(tmpAlgSequence,key)

""")
if 'RDO' in outputFiles:
    oFile.write("""
key = "StreamRDO"    
if key in _configs:
    StreamRDO = _getConfig( key )
else:
    StreamRDO = Algorithm( key )
""")
    oFile.write('StreamRDO.OutputFile = "%s"\n' % outputFiles['RDO'])
    oFile.write('pTmpStreamList.append(StreamRDO)\n')
if 'ESD' in outputFiles:
    oFile.write("""
key = "StreamESD"    
if key in _configs:
    StreamESD = _getConfig( key )
else:
    StreamESD = Algorithm( key )
""")
    oFile.write('StreamESD.OutputFile = "%s"\n' % outputFiles['ESD'])
    oFile.write('pTmpStreamList.append(StreamESD)\n')
if 'AOD' in outputFiles:
    oFile.write("""
key = "StreamAOD"    
if key in _configs:
    StreamAOD = _getConfig( key )
else:
    StreamAOD = Algorithm( key )
""")
    oFile.write('StreamAOD.OutputFile = "%s"\n' % outputFiles['AOD'])
    oFile.write('pTmpStreamList.append(StreamAOD)\n')
if 'TAG' in outputFiles:
    oFile.write("""
key = "StreamTAG"    
if key in _configs:
    StreamTAG = _getConfig( key )
else:
    StreamTAG = Algorithm( key )
""")
    oFile.write('StreamTAG.OutputCollection = "%s"\n' % re.sub('\.root\.*\d*$','',outputFiles['TAG']))
if 'AANT' in outputFiles:
    firstFlag = True
    oFile.write('THistSvc = _Service ( "THistSvc" )\n')
    sNameList = []
    for aName,sName,fName in outputFiles['AANT']:
        if not sName in sNameList:
            sNameList.append(sName)    
            if firstFlag:
                firstFlag = False
                oFile.write('THistSvc.Output = ["%s DATAFILE=\'%s\' OPT=\'UPDATE\'"]\n' % (sName,fName))
            else:
                oFile.write('THistSvc.Output += ["%s DATAFILE=\'%s\' OPT=\'UPDATE\'"]\n' % (sName,fName))            
        oFile.write("""
key = "%s"
if key in _configs:
    AANTupleStream = _getConfig( key )
else:
    AANTupleStream = Algorithm( key )
""" % aName)
        oFile.write('AANTupleStream.StreamName = "%s"\n' % sName)
        oFile.write('AANTupleStream.OutputName = "%s"\n' % fName)        
    if 'THIST' in outputFiles:
        for sName,fName in outputFiles['THIST']:
            oFile.write('THistSvc.Output += ["%s DATAFILE=\'%s\' OPT=\'UPDATE\'"]\n' % (sName,fName))
else:
    if 'THIST' in outputFiles:
        oFile.write('THistSvc = _Service ( "THistSvc" )\n')
        firstFlag = True
        for sName,fName in outputFiles['THIST']:
            if firstFlag:
                firstFlag = False
                oFile.write('THistSvc.Output = ["%s DATAFILE=\'%s\' OPT=\'UPDATE\'"]\n' % (sName,fName))
            else:
                oFile.write('THistSvc.Output+= ["%s DATAFILE=\'%s\' OPT=\'UPDATE\'"]\n' % (sName,fName))
if 'Stream1' in outputFiles:
    oFile.write("""
key = "Stream1"    
if key in _configs:
    Stream1 = _getConfig( key )
else:
    try:
        Stream1 = getattr(theApp._streams,key)
    except:
        Stream1 = Algorithm( key )
""")
    oFile.write('Stream1.OutputFile = "%s"\n' % outputFiles['Stream1'])
    oFile.write('pTmpStreamList.append(Stream1)\n')
if 'Stream2' in outputFiles:
    oFile.write("""
key = "Stream2"    
if key in _configs:
    Stream2 = _getConfig( key )
else:
    try:
        Stream2 = getattr(theApp._streams,key)
    except:
        Stream2 = Algorithm( key )
""")
    oFile.write('Stream2.OutputFile = "%s"\n' % outputFiles['Stream2'])
    oFile.write('pTmpStreamList.append(Stream2)\n')
    oFile.write("""
key = "%s_FH" % key
Stream2_FH = None
if key in _configs:
    Stream2_FH = _getConfig( key )
else:
    try:
        Stream2_FH = getattr(theApp._streams,key)
    except:
        pass
""")
    oFile.write("""
if Stream2_FH != None:
    Stream2_FH.OutputFile = "%s"
""" % outputFiles['Stream2'])

if 'StreamG' in outputFiles:
    for stName,stFileName in outputFiles['StreamG']:
        oFile.write("""
key = "%s"    
if key in _configs:
    StreamX = _getConfig( key )
else:
    try:
        StreamX = getattr(theApp._streams,key)
    except:
        StreamX = Algorithm( key )
""" % stName)
        oFile.write('StreamX.OutputFile = "%s"\n' % stFileName)
        oFile.write('pTmpStreamList.append(StreamX)\n')    
if 'Meta' in outputFiles:
    for stName,stFileName in outputFiles['Meta']:
        oFile.write("""
key = "%s"    
if key in _configs:
    StreamX = _getConfig( key )
else:
    try:
        StreamX = getattr(theApp._streams,key)
    except:
        StreamX = Algorithm( key )
""" % stName)
        oFile.write('StreamX.OutputFile = "ROOTTREE:%s"\n' % stFileName)

if 'UserData' in outputFiles:
    for stFileName in outputFiles['UserData']:
        oFile.write("""
try:
    # try new style
    userDataSvc = None
    try:
        for typeNameExtSvc in theApp.ExtSvc:
            if typeNameExtSvc.startswith('UserDataSvc/'):
                nameExtSvc = typeNameExtSvc.split('/')[-1]
                userDataSvc = getattr(theApp.serviceMgr(),nameExtSvc)
    except:
        pass
    # use old style
    if userDataSvc == None: 
        userDataSvc = _Service('UserDataSvc')
    # delete existing stream
    try:
       THistSvc = _Service ('THistSvc')
       tmpStreams = tuple(THistSvc.Output)
       newStreams = []
       for tmpStream in tmpStreams:
            # skip userstream since it is set by userDataSvc.OutputStream later
            if not userDataSvc.name() in tmpStream.split()[0]: 
                newStreams.append(tmpStream)
       THistSvc.Output = newStreams
    except:
        pass
    for tmpStream in pTmpStreamList:
        try:
            if tmpStream.OutputFile == '%s':
                userDataSvc.OutputStream = tmpStream
                break
        except:
            pass
except:
    pass
""" % stFileName)
    
uniqueTag = commands.getoutput('uuidgen 2>/dev/null')
if 'BS' in outputFiles:
    oFile.write('ByteStreamEventStorageOutputSvc = _Service("ByteStreamEventStorageOutputSvc")\n')
    oFile.write('ByteStreamEventStorageOutputSvc.FileTag = "%s"\n' % uniqueTag)
    oFile.write("""
try:
    ByteStreamEventStorageOutputSvc.AppName = "%s"
except:
    pass
""" % uniqueTag)    
    oFile.write('ByteStreamEventStorageOutputSvc.OutputDirectory = "./"\n')
if fragmentJobO != "":
    oFile.write('%s\n' % fragmentJobO)

# event picking
if eventPickTxt != '':
    epRunEvtList = []
    epFH = open(eventPickTxt)
    iepNum = 0
    for epLine in epFH:
        items = epLine.split()
        if len(items) != 2:
            continue
        # check range    
        epSkipFlag = False    
        if eventPickNum > 0:
            if iepNum < eventPickSt:
                epSkipFlag = True
            if iepNum >= eventPickSt+eventPickNum:
                epSkipFlag = True
        iepNum += 1
        if epSkipFlag:
            continue
        # append        
        epRunEvtList.append((long(items[0]),long(items[1])))
    oFile.write("""
from AthenaCommon.AlgSequence import AthSequencer
seq = AthSequencer('AthFilterSeq')
from GaudiSequencer.PyComps import PyEvtFilter
seq += PyEvtFilter(
    'alg',
    evt_info='',
    )
seq.alg.evt_list = %s
seq.alg.filter_policy = 'accept'
for tmpStream in theApp._streams.getAllChildren():
    fullName = tmpStream.getFullName()
    if fullName.split('/')[0] == 'AthenaOutputStream':
        tmpStream.AcceptAlgs = [seq.alg.name()]
""" % str(epRunEvtList))

# FileStager
if useFileStager and directIn:
    try:
        print "=== preparation for FileStager ==="
        # remove log just in case
        commands.getoutput('rm -f %s.log' % pCopyToolName)
        # get FileStager jobO    
        fileStagerDir = 'FileStager'
        if newPrefix.startswith('srm:'):
            fileStagerJobO = 'input_FileStager.py'
        else:    
            fileStagerJobO = 'input_FileStagerRFCP.py'
        print commands.getoutput('rm -f %s;get_files %s/%s' % (fileStagerJobO,fileStagerDir,fileStagerJobO))
        # create dummy jobOs to disable FileStager which is already included in user's jobO
        commands.getoutput('mkdir -p %s' % fileStagerDir)
        for tmpFsJobO in ['input_FileStager.py','input_FileStagerRFCP.py']:
            commands.getoutput('rm -f %s/%s' % (fileStagerDir,tmpFsJobO))
            commands.getoutput('touch %s/%s' % (fileStagerDir,tmpFsJobO))
        # tweak jobO for dcap/xrootd
        fsStrs = ''
        fsFH = open(fileStagerJobO)
        copyToolCpCommand = ''
        for tmpLine in fsFH:
            tmpMatch = re.search('(.*)stagetool.CpCommand\s*=\s*".+"',tmpLine)
            if tmpMatch != None:
                # replace copy command
                if copyTool == 'lsm':
                    # local site mover
                    fsStrs += '%sstagetool.CpCommand="lsm-get"\n' % tmpMatch.group(1)
                elif newPrefix.startswith('root:'):
                    # xrootd
                    copyToolCpCommand = "xrdcp"
                    fsStrs += '%sstagetool.CpCommand="./%s"\n' % (tmpMatch.group(1),pCopyToolName)
                elif newPrefix.startswith('dcap:') or newPrefix.startswith('dcache:') or newPrefix.startswith('gsidcap:'):
                    # dCache
                    copyToolCpCommand = "dccp"
                    fsStrs += '%sstagetool.CpCommand="./%s"\n' % (tmpMatch.group(1),pCopyToolName)
                elif newPrefix.startswith('rfio:') or newPrefix.startswith('castor:'):
                    # rfio/castor
                    copyToolCpCommand = "rfcp"
                    fsStrs += '%sstagetool.CpCommand="./%s"\n' % (tmpMatch.group(1),pCopyToolName)
                elif newPrefix.startswith('/'):
                    # NAS
                    copyToolCpCommand = "cp"
                    fsStrs += '%sstagetool.CpCommand="./%s"\n' % (tmpMatch.group(1),pCopyToolName)
                else:
                    # keep original
                    fsStrs += tmpLine                
                fsStrs += '%sstagetool.tmpDir = "%s/"\n' % (tmpMatch.group(1),currentDir)
            else:
                fsStrs += tmpLine
        fsStrs += """
try:
    thejob.FileStager.KeepLogfiles = True
except:
    pass
try:
    thejob.FileStager.VerboseStager = True
except:
    pass
try:
    thejob.FileStager.CpCommand = stagetool.CpCommand
except:
    pass
try:
    thejob.FileStager.CpArguments = stagetool.CpArguments
except:
    pass
"""
        fsFH.close()
        # create new jobO
        fsFH = open(fileStagerJobO,'w')
        fsFH.write(fsStrs)
        fsFH.close()
        # create copy wrapper
        copyToolFH = open(pCopyToolName,'w')
        copyToolFH.write(pCopyToolStr)
        copyToolFH.close()
        commands.getoutput('chmod +x %s' % pCopyToolName)
        # create sample file
        sampleFileName = 'sample_%s.def' % commands.getoutput('uuidgen 2>/dev/null')
        sampleFH = open(sampleFileName,'w')
        sampleFH.write('TITLE: title\nFLAGS: GridCopy=1\n')
        metaDataForFH = {}
        for tmpInputName in inputFiles:
            tmpCheckSum = None
            # look for GUID
            for tmpGUIDforMeta,tmpPFNforMeta in directPfnMap.iteritems():
                if tmpPFNforMeta == tmpInputName:
                    # get checksum
                    if tmpGUIDforMeta in directMetaMap and 'checksum' in directMetaMap[tmpGUIDforMeta]:
                        tmpCheckSum = directMetaMap[tmpGUIDforMeta]['checksum']
                    break    
            if tmpInputName.startswith('dcache:'):
                # remove prefix 
                tmpInputName = re.sub('dcache:','',tmpInputName)
            # append checksum
            if tmpCheckSum != None:
                metaDataForFH[tmpInputName] = tmpCheckSum
            # add prefix
            tmpInputName = 'gridcopy://%s\n' % tmpInputName
            # write
            sampleFH.write(tmpInputName)
        sampleFH.close()
        # create config for copyTool
        copytoolConfFH = open('%sConf.py' % pCopyToolName,'w')
        copytoolConfFH.write('com = "%s"\n' % copyToolCpCommand)
        copytoolConfFH.write('metaMap = %s\n' % str(metaDataForFH))
        copytoolConfFH.close()
        print "----> jobO"
        print commands.getoutput('cat %s' % fileStagerJobO)
        print "----> sample"        
        print commands.getoutput('cat %s' % sampleFileName)
        print "----> config"
        print commands.getoutput('cat %sConf.py' % pCopyToolName)
    except:
        type,value,traceBack = sys.exc_info()
        print 'ERROR : failed to setup FileStager %s - %s' % (type,value)
    # modify PoolFC.xml
    pfcName = 'PoolFileCatalog.xml'
    try:
        pFile = open(pfcName)
        pFileStr = ''
        for tmpLine in pFile:
            pFileStr += tmpLine 
        pFile.close()
        for tmpInputName in inputFiles:
            # replace file name
            baseName = '%s/tcf_%s' % (currentDir,tmpInputName.split('/')[-1])
            pFileStr = re.sub(tmpInputName,baseName,pFileStr)
        # overwrite
        pFile = open(pfcName,'w')
        pFile.write(pFileStr)
        pFile.close()
    except:
        pass
    # add jobO
    oFile.write("""
try:
    del sampleList
except:
    pass
sampleFile = '%s'
doStaging = True
include ('%s')
""" % (sampleFileName,fileStagerJobO))
        
oFile.close()

# overwrite EventSelectorAthenaPool.InputCollections and AthenaCommon.AthenaCommonFlags.FilesInput for jobO level metadata extraction
preOpt = 'pre_' + commands.getoutput('uuidgen 2>/dev/null') + '.py'
oFile = open(preOpt,'w')
if len(inputFiles) != 0 and mcData == '':
    if not byteStream:
        oFile.write("""      
try:
    from EventSelectorAthenaPool.EventSelectorAthenaPoolConf import EventSelectorAthenaPool
    orig_ESAP__getattribute =  EventSelectorAthenaPool.__getattribute__

    def _dummy(self,attr):
        if attr == 'InputCollections':
            return %s
        else:
            return orig_ESAP__getattribute(self,attr)

    EventSelectorAthenaPool.__getattribute__ = _dummy
    print 'Overwrite InputCollections'
    print EventSelectorAthenaPool.InputCollections
except:
    try:
        EventSelectorAthenaPool.__getattribute__ = orig_ESAP__getattribute
    except:
        pass
""" % inputFiles)
    oFile.write("""      
try:
    import AthenaCommon.AthenaCommonFlags

    def _dummyFilesInput(*argv):
        return %s

    AthenaCommon.AthenaCommonFlags.FilesInput.__call__ = _dummyFilesInput
except:
    pass

try:
    import AthenaCommon.AthenaCommonFlags

    def _dummyGet_Value(*argv):
        return %s

    for tmpAttr in dir (AthenaCommon.AthenaCommonFlags):
        import re
        if re.search('^(Pool|BS).*Input$',tmpAttr) != None:
            try:
                getattr(AthenaCommon.AthenaCommonFlags,tmpAttr).get_Value = _dummyGet_Value
            except:
                pass
except:
    pass
""" % (inputFiles,inputFiles))
# filter for verbose expansion
try:
    if athenaVer >= "15.0.0":
        oFile.write("""
try:
    from AthenaCommon.Include import excludeTracePattern
    excludeTracePattern.append('*/CLIDComps/clidGenerator.py')
    excludeTracePattern.append('*/PyUtils/decorator.py')
    excludeTracePattern.append('*/PyUtils/Decorators.py')
    excludeTracePattern.append('*/PyUtils/Helper*.py')
except:
    pass
""")
except:
    pass
# for SummarySvc
oFile.write("""      
try:
    from AthenaServices.SummarySvc import *
    useAthenaSummarySvc()
except:
    pass
""")
                
oFile.close()

# dump

print "=== pre jobO ==="
oFile = open(preOpt)
lines = ''
for line in oFile:
    lines += line
print lines
oFile.close()
print

print "=== post jobO ==="
oFile = open(postOpt)
lines = ''
for line in oFile:
    lines += line
print lines
oFile.close()

# replace theApp.initialize when using theApp.nextEvent
if useNextEvent:
    initOpt = 'init_' + commands.getoutput('uuidgen 2>/dev/null') + '.py'
    initFile = open(initOpt,'w')
    initFile.write("""
origTheAppinitialize = theApp.initialize                   
def fakeTheAppinitialize():
    include('%s')
    origTheAppinitialize()
theApp.initialize = fakeTheAppinitialize    
""" % postOpt)
    initFile.close()
    
    print "=== init jobO ==="
    iFile = open(initOpt)
    lines = ''
    for line in iFile:
        lines += line
    print lines
    iFile.close()

    # modify jobO
    print "=== change jobO ==="
    newJobO = ''
    startPy = False
    for item in jobO.split():
        if (not startPy) and item.endswith('.py'):
            newJobO += (" " + initOpt)
            startPy = True
        newJobO += (" " + item)    
    print "  Old : " + jobO
    print "  New : " + newJobO
    jobO = newJobO
    
# modify macro for ARA
if runAra and len(inputFiles) != 0:
   try:
       # look for .C or .py
       for tmpIdx in range(len(jobO.split())):
           tmpRevIdx = -(tmpIdx+1)
           tmpMacroName = jobO.split()[tmpRevIdx]
           if (not tmpMacroName.startswith('-')) and \
               (tmpMacroName.endswith('.py') or tmpMacroName.endswith('.C')):
               break
       if tmpMacroName.endswith('.C'):
           # look for TPython::Exec
           tmpMacro = open(tmpMacroName)           
           for line in tmpMacro:
               match = re.search('TPython::Exec\s*\(.*execfile[^\'\"]+[\'\"]([^\'\"]+)',line)
               if match != None:
                   tmpMacroName = match.group(1)
                   break
           # close
           tmpMacro.close()
       # check startup file
       if not tmpMacroName.endswith('.py'):
           raise RuntimeError('startup file (%s) needs to be .py for ARA' % tmpMacroName)
       newMacroName = tmpMacroName+'.new'
       tmpMacro = open(tmpMacroName)
       newMacro = open(newMacroName,'w')       
       # look for chain building sector
       inputSector = False
       for line in tmpMacro:
           # remove comments
           if re.search('^\s*#',line) != None:
               continue
           # look for TChainROOTAccess       
           match = re.search('(\s*).*ROOT.AthenaROOTAccess.TChainROOTAccess',line)
           if match != None:
               inputSector = True
               newMacro.write(line)
               # execfile to set Chain
               newMacro.write('%sexecfile("%s")\n' % (match.group(1),postOpt))
               continue
           # look for makeTree    
           match = re.search('AthenaROOTAccess.transientTree.makeTree',line)
           if match != None:
               inputSector = False
               newMacro.write(line)
               continue
           # write out normal sectors
           if not inputSector:
               newMacro.write(line)
       tmpMacro.close()
       newMacro.close()
       # dump
       print "=== old %s ===" % tmpMacroName
       tmpFile = open(tmpMacroName)
       lines = ''
       for line in tmpFile:
           lines += line
       print lines
       tmpFile.close()
       print "=== new %s ===" % newMacroName
       tmpFile = open(newMacroName)
       lines = ''
       for line in tmpFile:
           lines += line
       print lines
       tmpFile.close()
       # rename
       commands.getoutput('mv %s %s' % (newMacroName,tmpMacroName))
   except:
       type, value, traceBack = sys.exc_info()
       print 'ERROR: failed to modify %s : %s - %s' % (tmpMacroName,type,value)
       sys.exit(EC_ARA)

# get PDGTABLE.MeV
commands.getoutput('get_files PDGTABLE.MeV')

# temporary output to avoid MemeoryError
tmpOutput = 'tmp.stdout.%s' % commands.getoutput('uuidgen 2>/dev/null')
tmpStderr = 'tmp.stderr.%s' % commands.getoutput('uuidgen 2>/dev/null')

# append workdir to CMTPATH
env = 'CMTPATH=%s:$CMTPATH' % workDir
# construct command
com  = 'export %s;' % env
# local RAC
if ('ATLAS_CONDDB' not in os.environ) or os.environ['ATLAS_CONDDB']=='to.be.set':
    if 'OSG_HOSTNAME' in os.environ:
        com += 'export ATLAS_CONDDB=%s;' % os.environ['OSG_HOSTNAME']
    elif 'GLOBUS_CE' in os.environ:
        tmpCE = os.environ['GLOBUS_CE'].split('/')[0]
        # remove port number
        tmpCE = re.sub(':\d+$','',tmpCE)
        com += 'export ATLAS_CONDDB=%s;' % tmpCE
    elif 'PBS_O_HOST' in os.environ:
        com += 'export ATLAS_CONDDB=%s;' % os.environ['PBS_O_HOST']    
com += 'cd %s;' % cmtDir
com += 'echo "use AtlasPolicy AtlasPolicy-*" > requirements;'
com += 'cmt config;'
com += 'source ./setup.sh;'
com += 'export TestArea=%s;' % workDir
com += 'cd -;env;'
if (not runTrf) and dbrFile == '':
    # run Athena
    com += 'athena.py '
    if codeTrace:
        com += '-s '
    com += '%s %s %s' % (preOpt,jobO,postOpt)
elif dbrFile != '':
    # run setup.py and athena.py in a python
    tmpTrfName = 'trf.%s.py' % commands.getoutput('uuidgen 2>/dev/null')
    tmpTrfFile = open(tmpTrfName,'w')
    tmpTrfFile.write(dbrSetupStr)
    tmpTrfFile.write('import sys\nstatus=os.system("""athena.py ')
    if codeTrace:
        tmpTrfFile.write('-s ')
    tmpTrfFile.write('%s %s %s""")\n' % (preOpt,jobO,postOpt))
    tmpTrfFile.write('status %= 255\nsys.exit(status)\n\n')
    tmpTrfFile.close()
    com += 'cat %s;python -u %s' % (tmpTrfName,tmpTrfName)
else:
    # run transformation
    com += '%s' % jobO

print    
print "=== execute ==="
print com
# run athena
if not debugFlag:
    if liveLog != '':
        # create empty log
        commands.getstatusoutput('cat /dev/null > %s' % tmpOutput)
        # start watcher
        logWatcher = LogWatcher(tmpOutput,liveLog)
        logWatcher.start()
    # write stdout to tmp file
    com += ' > %s 2> %s' % (tmpOutput,tmpStderr)
    status,out = commands.getstatusoutput(com)
    print out
    statusChanged = False
    try:
        tmpOutFile = open(tmpOutput)
        for line in tmpOutFile:
            print line[:-1]
            # set status=0 for AcerMC
            if re.search('ACERMC TERMINATES NORMALY: NO MORE EVENTS IN FILE',line) != None:
                status = 0
                statusChanged = True
        tmpOutFile.close()
    except:
        pass
    if statusChanged:
        print "\n\nStatusCode was overwritten for AcerMC\n"
    try:
        tmpErrFile = open(tmpStderr)
        for line in tmpErrFile:
            print line[:-1]
        tmpErrFile.close()
    except:
        pass
    # print 'sh: line 1:  8278 Aborted'
    try:
        if status != 0:
            print out.split('\n')[-1]
    except:
        pass
    if liveLog != '':        
        # terminate watcher
        logWatcher.terminate()
        logWatcher.join()
else:
    status = os.system(com)

if useFileStager and directIn:
    print
    print "=== FileStager log ==="
    print commands.getoutput('cat %s.log' % pCopyToolName)
    print commands.getoutput('cat %s/tcf_*stage.err 2>/dev/null' % currentDir)

print
print "=== list in run dir ==="    
print commands.getoutput('ls -l')

# rename or archive iROOT files
if 'IROOT' in outputFiles:
    for iROOT in outputFiles['IROOT']:
        if iROOT[0].find('*') != -1:
            # archive *
            commands.getoutput('tar cvfz %s %s' % (iROOT[-1],iROOT[0]))
        else:
            # rename 
            commands.getoutput('mv %s %s' % iROOT)
        # modify PoolFC.xml
        pfcName = 'PoolFileCatalog.xml'
        try:
            pLines = ''
            pFile = open(pfcName)
            for line in pFile:
                # replace file name
                line = re.sub('"%s"' % iROOT[0],'"%s"' % iROOT[-1],line)
                pLines += line
            pFile.close()
            # overwrite
            pFile = open(pfcName,'w')
            pFile.write(pLines)
            pFile.close()
        except:
            pass

# rename TAG files
if 'TAG' in outputFiles:
    woAttrNr = re.sub('\.\d+$','',outputFiles['TAG'])
    if woAttrNr != outputFiles['TAG']:
        print commands.getoutput('mv %s %s' % (woAttrNr,outputFiles['TAG']))
    # since 13.0.30 StreamTAG doesn't append .root automatically
    woRootAttrNr = re.sub('\.root\.*\d*$','',outputFiles['TAG'])
    if woRootAttrNr != outputFiles['TAG']:
        print commands.getoutput('mv %s %s' % (woRootAttrNr,outputFiles['TAG']))

# rename BS file
if 'BS' in outputFiles:
    bsS,bsO = commands.getstatusoutput('mv daq.%s* %s' % (uniqueTag,outputFiles['BS']))
    print bsS,bsO
    if bsS != 0:
        print commands.getstatusoutput('mv data_test.*%s* %s' % (uniqueTag,outputFiles['BS']))    
    
# copy results
for file in outputFiles.values():
    if not isinstance(file,types.StringType):
        # for AANT
        for aaT in file:
            commands.getoutput('mv %s %s' % (aaT[-1],currentDir))
    else:
        commands.getoutput('mv %s %s' % (file,currentDir))

# copy PoolFC.xml
commands.getoutput('mv -f PoolFileCatalog.xml %s' % currentDir)

# copy AthSummary.txt
commands.getoutput('mv -f AthSummary.txt %s' % currentDir)

# go back to current dir
os.chdir(currentDir)

print
print "=== list in top dir ==="    
print commands.getoutput('pwd')
print commands.getoutput('ls -l')

# remove work dir
if not debugFlag:
    commands.getoutput('rm -rf %s' % workDir)

# return
print
print "=== result ==="    
if status:
    print "execute script: Running athena failed : %d" % status
    sys.exit(EC_AthenaFail)
else:
    print "execute script: Running athena was successful"
    sys.exit(0)



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 17/01/2014
################################################################################
"""@package ND280Dataset
Ganga module to control the input files.
It relies on the basic ND280Dataset class which is used through inheritance in other more specific classes.
"""

from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger

import os, re, fnmatch
import commands

logger = getLogger()

class ND280Dataset(Dataset):
    """
    Base class for ND280 Datasets.
        dataset = ND280Dataset()

    You can define the list of files by hand directly if you want:
        dataset.names = ['/path/to/the/files/Input1.root','/path/to/the/files/Input2.root']
    """
    
    _schema = Schema(Version(1,0), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path'),
    })
    
    _category = 'datasets'
    _name = 'ND280Dataset'
    _hidden = 1

    def __init__(self):
        super(ND280Dataset, self).__init__()
        
    def get_dataset_from_list(self,list_file):
       """Get the dataset files as listed in a text file."""

       logger.info('Reading list file %s ...',list_file)

       if not os.path.exists(list_file):
           logger.error('File %s does not exist',list_file)
           return

       f = open( list_file )
       for ln in f.readlines():

           # split the directory from the file and call get_dataset
           if os.path.isdir(ln.strip()):
               self.get_dataset( ln.strip() )
           else:
               self.get_dataset( os.path.dirname(ln.strip()), os.path.basename( ln.strip() ) )
        
    def set_dataset_into_list(self,list_file):
       """Write the dataset files as a list in a text file."""

       logger.info('Writing list file %s ...',list_file)

       if not len(self.names):
         logger.error('No input files found. The dataset is empty.')
         return 0

       try:
         fileList = open( list_file, 'w')
       except IOError:
         logger.error('File %s cannot be created',list_file)
         return 0

       for inFile in self.names:
         fileList.write(inFile+'\n')

       fileList.close()
       return 1

    def get_dataset(self,directory,filter=None):
        """Get the list of files in the dataset's directory.
        This is not implemented in ND280Dataset but is defined in each class inheriting from ND280Dataset."""
      
        raise NotImplementedError


    def get_dataset_filenames(self):
        """Simply returns a python list containing all the filenames in this ND280Dataset.
        """
        return self.names

    def set_dataset_filenames(self,list_file):
       """Copy the list of files given in input as the list of files in the ND280Dataset.
       NOTE: This will not append the input list to the existing file list but instead replace it."""

       logger.info('Writing list file %s ...',list_file)

       self.names=list_file

    def get_filenames(app):
        """Retrieve the file names starting from an application object"""
      
        job=app._getRoot()
        if not job:
            logger.warning('Application object is not associated to a job.')
            return []
         
        # Jobs without inputdata are allowed
        if not job.inputdata: return []
      
        import re
        classnamematch = re.match(r"ND280.*Dataset", job.inputdata._name)
        if not classnamematch:
            logger.warning('Dataset is not a class inheriting from ND280Dataset.')
            return []

        return job.inputdata.names
         
    get_filenames=staticmethod(get_filenames)


class ND280LocalDataset(ND280Dataset):

    """ND280LocalDataset manages files located in a local directory."""
    
    _schema = Schema(Version(1,1), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path'),
    })
    
    _category = 'datasets'
    _name = 'ND280LocalDataset'

    _exportmethods = ['get_dataset', 'get_dataset_filenames', 'get_dataset_from_list', 'get_raw_from_list', 'get_kin_range','set_dataset_into_list', 'set_dataset_filenames']

    _GUIPrefs = [ { 'attribute' : 'names',  'attribute' :  'String_List' } ]
   
    def __init__(self):
        super(ND280LocalDataset, self).__init__()


    def get_dataset(self,directory,filter=None):
       """Get the list of files in the local directory given.
       You can provide a wildcard filter such as 'oa_*.root'"""
      
       logger.info('Reading %s ...',directory)


       if not os.path.isdir(directory):
           logger.error('Path %s is no directory',directory)
           return

       directory = os.path.abspath(directory)
       if filter:
           new_names = [ os.path.join(directory,name) for name in fnmatch.filter(os.listdir(directory),filter) ]
       else:
           new_names = [ os.path.join(directory,name) for name in os.listdir(directory) ]

       self.names.extend( new_names )

       self._setDirty()

    def get_raw_from_list(self,prfx,list_file):
       """Get the dataset of raw files as listed in a text file as run/subrun combinations."""

       logger.info('Reading list file %s ...',list_file)

       if not os.path.isdir(prfx):
           logger.error('Directory $s does not exist',prfx)
           return

       if not os.path.exists(list_file):
           logger.error('File %s does not exist',list_file)
           return

       f = open( list_file )
       for ln in f.readlines():

           chunks = ln.split()
           run = "%08d" % int(chunks[0])
           sub = "%04d" % int(chunks[1])
           rang = run[:5]+'000_'+run[:5]+'999'
           file = 'nd280_'+run+'_'+sub+'.daq.mid.gz'
           self.get_dataset( os.path.join(prfx,rang), file)

    def get_kin_range(self,fr,to):
        """Get the dataset of kin file numbers"""

        logger.info('Producing a list of kin file numbers in the range from %s to %s.',fr,to)

        self.names.extend([j for j in range(fr,to+1)])


class ND280DCacheDataset(ND280Dataset):

    """ND280 local datasets manages files located in a directory on a DCache server.
    By default, the configured server is TRIUMF.
    And currently the only configured server is TRIUMF but later you will be able to use another server:
      dataset.server = 'TRIUMF'
    """
    
    _schema = Schema(Version(1,0), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path to get the file on the server, i.e. "dcap://the/path/thefile'),
        'server': SimpleItem(defvalue = "TRIUMF", doc='Name of the dcache server used'),
    })
    
    _name = 'ND280DCacheDataset'

    _exportmethods = ['get_dataset', 'get_dataset_filenames', 'get_dataset_from_list', 'set_dataset_into_list', 'set_dataset_filenames' ]

    _GUIPrefs = [ { 'attribute' : 'names',  'attribute' :  'server',  'widget' : 'String_List' } ]

    _commandstr = getConfig('ND280')['ND280DCacheDatasetCommandStr']
    _filebasepath = getConfig('ND280')['ND280DCacheDatasetFileBasePath']

    def __init__(self):
        super(ND280DCacheDataset, self).__init__()


    def get_dataset(self,directory,filter=None):
      """Get the list of files in the directory on the dCache server.
      You can provide a wildcard filter such as 'oa_*.root'"""

      if not self.server in self._commandstr.keys():
          logger.error('DCache server %s is unknown.', self.server)
          return

      logger.info('Reading %s ...',directory)

      command = self._commandstr[self.server] % directory
      rawoutput = commands.getoutput(command);
      allfiles = rawoutput.split("\n")

      # TODO: return directory error when curl isn't happy
      #if not os.path.isdir(directory):
      #    logger.error('Path %s is no directory',directory)
      #    return

      fullpath = os.path.join(self._filebasepath[self.server], directory )
      if filter:
          new_names = [ os.path.join(fullpath,name) for name in fnmatch.filter(allfiles,filter) ]
      else:
          new_names = [ os.path.join(fullpath,name) for name in allfiles ]

      self.names.extend( new_names )

      self._setDirty()



from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Schema import *

# Test Dataset

class TestDataset(Dataset):
    _schema = Schema(Version(1,0), {'files':FileItem(defvalue="",sequence=1)})
    _category = 'datasets'
    _name = "TestDataset"

    def __init__(self):
        super(TestDataset, self).__init__()

    def isEmpty(self):
        return bool(self.files)

# a dataset which looks like a list
class TestListViewDataset(Dataset):
    _schema = Schema(Version(1,0), {'files':FileItem(defvalue="",sequence=1)})
    _category = 'datasets'
    _name = "TestListViewDataset"

    def __init__(self):
        super(TestListViewDataset, self).__init__()

    def isEmpty(self):
        return bool(self.files)

    def _object_filter__get__(self,obj):
        #return self.files
        return obj

from Ganga.GPIDev.Base.Filters import allComponentFilters

def list_assignment_shortcut(v,item):
    if type(v) is type([]):
        # use proxy class to enable all user conversions on the value itself
        # but return the implementation object (not proxy)
        d = TestListViewDataset._proxyClass()
        d.files = v
        return d._impl
    else:
        return None
        
allComponentFilters['datasets'] = list_assignment_shortcut
    



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: AthenaMCLCGRTHandler.py,v 1.10.2.3 2006/11/22 14:20:53 elmsheus Exp 
###############################################################################
# AthenaMC LCG Runtime Handler
#

import os, string, commands, re, pwd

import imp,xml.dom.minidom,urllib,random

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *
from Ganga.Core import FileWorkspace
from Ganga.Core.exceptions import ApplicationConfigurationError

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.logging import getLogger

from Ganga.Lib.LCG import LCGRequirements, LCGJobConfig
from GangaAtlas.Lib.AtlasLCGRequirements import AtlasLCGRequirements
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

from GangaAtlas.Lib.ATLASDataset import DQ2Dataset
from GangaAtlas.Lib.ATLASDataset import DQ2OutputDataset
from GangaAtlas.Lib.AthenaMC.AthenaMCDatasets import extractFileNumber, matchFile


from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

_defaultSite='CERN-PROD_SCRATCHDISK'

class AthenaMCLCGRTHandler(IRuntimeHandler):
    """Athena MC LCG Runtime Handler"""

    def sortSites(self,insites,outsite):
        inlist=insites.split(" ")
        imax=outsite.find("_")
        outsite=outsite[:imax]
        newlist=[]
        for site in inlist:
            if site.find(outsite)>-1:
                newlist.insert(0,site)
            else:
                newlist.append(site)
        return string.join(newlist," ")
            
    def master_prepare(self,app,appmasterconfig):
        if app.siteroot: 
            os.environ["SITEROOT"]=app.siteroot
        os.environ["CMTSITE"]=app.cmtsite

        job = app._getParent()
        if job.backend._name in ["Local","PBS"]:
            if app.dryrun:
                os.environ["SITEROOT"]  = "NONE"
                os.environ["CMTSITE"]  = "NONE"
            try:
                assert "SITEROOT" in os.environ
            except:
                raise ApplicationConfigurationError(None," ATLAS environment not defined")
                
            try:
                assert "CMTSITE" in os.environ
            except:
                raise ApplicationConfigurationError(None,"cmt not setup properly. Please check your ATLAS setup or run on the grid")
            
            if os.environ["CMTSITE"]=="CERN" and "AtlasVersion" in os.environ:
                logger.debug("Checking AtlasVersion: %s and selected atlas release %s" % (os.environ["AtlasVersion"],app.atlas_rel))
                try:
                    assert app.atlas_rel==os.environ["AtlasVersion"]
                except:
                    logger.error("Mismatching atlas release. Local setup is %s, resetting requested atlas release to local value." % os.environ["AtlasVersion"])
                    app.atlas_release=os.environ["AtlasVersion"]
                    app.atlas_rel==os.environ["AtlasVersion"]
            elif "ATLAS_RELEASE" in os.environ:
                logger.debug("Checking ATLAS_RELEASE: %s and selected atlas release %s" % (os.environ["ATLAS_RELEASE"],app.atlas_rel))
                try:
                    assert app.atlas_rel==os.environ["ATLAS_RELASE"]
                except:
                    logger.error("Mismatching atlas release. Local setup is %s, resetting requested atlas release to local value." % os.environ["ATLAS_RELEASE"])
                    app.atlas_rel=os.environ["ATLAS_RELEASE"]
            else:
                logger.warning("Could not compare requested release and local setup. Hope you are doing something sensible...")

                
        if job.backend._name=="LSF":
            try:
                assert "CMTSITE" in os.environ and os.environ["CMTSITE"]=="CERN"
            except:
                raise ApplicationConfigurationError(None,"Error, CERN ATLAS AFS environment not defined. Needed by LSF backend")

       
        environment={'T_LCG_GFAL_INFOSYS' :'atlas-bdii.cern.ch:2170'}

        trfopts=app.transflags
        # need to parse them to be able to pass them in an environment variable
        trfopts=trfopts.replace(" ","/W")
        trfopts=trfopts.replace("-","/F")
        
        trflags=trfopts
        if app.mode =="evgen":
            trflags="/Ft"
            if app.verbosity:
                trflags+="/W/Fl/W%s" % app.verbosity
        
        if trflags:
            environment["TRFLAGS"]=trflags

        # setting output site from input data if any.
        outsite,backup,outputlocation,backuplocation="","","",""
        logger.info("checking sites from input data: %s" % str(app.sites))

        # must distinguish running site (backend.requirements.sites) and output storage site (app.se_name)
        
        # matching with user's wishes (app.se_name or backend.requirements.sites)

        usersites=[]
        if len(job.backend.requirements.sites)>0:
            usersites=job.backend.requirements.sites
##        elif job.application.se_name and job.application.se_name != "none":
##            usersites=job.application.se_name.split(" ")
        logger.info("user selection: %s" % str(usersites))
            
        # select sites which are matching user's wishes, if any.
        selectedSites=app.sites
        if len(selectedSites)==0:
            selectedSites=usersites
        if len(usersites)>0 and len(app.sites)>0:
            selectedSites=job.inputdata.trimSites(usersites,app.sites)
        # evgen case (no input data-> app.sites=[])
        if len(app.sites)==0 and app.se_name and app.se_name != "none":
            selectedSites=app.se_name.split(" ")

        # This comes last: using surviving sites from matching process.
        if len(selectedSites)==0:
            try:
                assert len(usersites)==0
            except:
                raise ApplicationConfigurationError(None,"Could not find a match between input dataset locations: %s and your requested sites: %s. Please use a space token compatible with one of the input dataset locations (replace _XXXDISK or _XXXTAPE by _LOCALGROUPDISK or _SCRATCHDISK if necessary)" % (str(app.sites),str(usersites)))
            logger.warning("Failed to obtain processing site from input data, will use default value: CERN-PROD_SCRATCHDISK and submit production to CERN")
            selectedSites.append(_defaultSite)


        [outlfc,outsite,outputlocation]=job.outputdata.getDQ2Locations(selectedSites[0])
        if len(selectedSites)>1:
            [outlfc2,backup,backuplocation]=job.outputdata.getDQ2Locations(selectedSites[1])

        # app.se_name set: users wishes to get the output data written to another site than the one hosting the input.
        # One needs to ensure that this location is at least in the same cloud as the targetted processing site. This is done by insuring that the lfcs are the same.
        userSEs=[]
        outse=""
        if job.application.se_name and job.application.se_name != "none":
            userSEs=job.application.se_name.split(" ")
            # loop through userSEs until up to 2 valid sites are found...
            outse=""
            for SE in userSEs:
                [lfc,se,location]=job.outputdata.getDQ2Locations(SE)
                if lfc==outlfc:
                    if not outse:
                        outse=se # important to use outse and not outsite here, as outsite is used for selection of processing site.
                        # userSEs overrides outlfc,outputlocation, but not outsite as outsite is unfortunately used for choice of the processing site.
                        outputlocation=location
                    else:
                        outlfc2=lfc
                        backup=se
                        backuplocation=location
                        break
        # finally: if no backup location is defined at this point, enforce CERN-PROD_SCRATCHDISK as backup location
        if backup=="":
             [outlfc2,backup,backuplocation]=job.outputdata.getDQ2Locations(_defaultSite)
        
        logger.info("Final selection of output sites: %s , backup: %s" % (outsite,backup))
        try:
            assert outsite
        except:
            raise ApplicationConfigurationError(None,"Could not find suitable location for your output. Please subscribe your input dataset (if any) to a suitable location or change application.se_name to a suitable space token")


        # srmv2 sites special treatment: the space token has been prefixed to the outputlocation and must be removed now:
        imin=string.find(outputlocation,"token:")
        imax=string.find(outputlocation,"srm:")
        spacetoken=""
        if imin>-1 and imax>-1:
            spacetoken=outputlocation[imin+6:imax-1]
            outputlocation=outputlocation[imax:]
        # same treatment for backup location if any
        imin=string.find(backuplocation,"token:")
        imax=string.find(backuplocation,"srm:")
        bst=""
        if imin>-1 and imax>-1:
            bst=backuplocation[imin+6:imax-1]
            backuplocation=backuplocation[imax:]

        environment["OUTLFC"]=outlfc
        environment["OUTSITE"]=outsite
        if outse:
           environment["OUTSITE"]=outse # user's choice for output storage location overriding AthenaMC's.
           
        environment["OUTPUT_LOCATION"]=outputlocation
        if spacetoken:
            environment["SPACETOKEN"]=spacetoken
        if backup:
            environment["OUTLFC2"]=outlfc2
            environment["OUTSITE2"]=backup
            environment["OUTPUT_LOCATION2"]=backuplocation

        environment["PROD_RELEASE"]=app.prod_release

        # setting environment["BACKEND"]
        # Local, Condor become "batch". LSF becomes "batch" unless the inputdata is on castor (in this case, it becomes "castor")
        environment["BACKEND"]=job.backend._name
        environment["BACKEND_DATA"]=app.backend_inputdata
        if job.backend._name=="LSF" and len(app.turls.values())>0:
            turl=app.turls.values()[0]
            if string.find(turl,"castor")>-1:
                environment["BACKEND_DATA"]="castor"
            else:
                environment["BACKEND_DATA"]="batch"
        if job.backend._name in ["Local","Condor","PBS"]:
            environment["BACKEND_DATA"]="batch"
            environment["SITEROOT"]=os.environ["SITEROOT"]
            environment["CMTSITE"]=os.environ["CMTSITE"]

#       finalise environment


        # preparing input sandbox, output sandbox , environment vars and job requirements
        
        inputbox = [ 
            File(os.path.join(os.path.dirname(__file__),'setup-release.sh')),
            File(os.path.join(os.path.dirname(__file__),'stage-in.sh')),
            File(os.path.join(os.path.dirname(__file__),'stage-out.sh')),
            File(os.path.join(os.path.dirname(__file__),'adler32.py'))
        ]

        if os.path.exists(app.transform_archive):
            # must add a condition on size.
            inputbox += [ File(app.transform_archive) ]
        elif app.transform_archive:
            # tarball in local or remote web area.
            if string.find(app.transform_archive,"http")>=0:
                environment['TRANSFORM_ARCHIVE'] = "%s" % (app.transform_archive)
            else:
                myfile=os.path.basename(app.transform_archive)
                myfile="http://cern.ch/atlas-computing/links/kitsDirectory/Production/kits/"+myfile
                environment['TRANSFORM_ARCHIVE'] = "%s" % (myfile)
                
        
        if app.evgen_job_option and os.path.exists(app.evgen_job_option):
            # locally modified job option file to add to the input sand box
            inputbox += [ File(app.evgen_job_option) ]
            # need to strip the path away.
            self.evgen_job_option = app.evgen_job_option.split("/")[-1]
            environment['CUSTOM_JOB_OPTION'] = "%s" % (self.evgen_job_option)
        elif app.evgen_job_option:
            self.evgen_job_option = app.evgen_job_option
            
        # user area:
        if app.userarea : 
            inputbox.append(File(app.userarea))
            environment['USER_AREA']=os.path.basename(app.userarea)


        if (job.inputsandbox):
            for file in job.inputsandbox:
                inputbox += [ file ]


        outputbox = [ ]
        outputGUIDs='output_guids'
        outputLOCATION='output_location'
        outputbox.append( outputGUIDs )
        outputbox.append( outputLOCATION )
        outputbox.append( 'output_data' )
        if (job.outputsandbox):
            for file in job.outputsandbox:
                outputbox += [ file ]

        # switch JobTransforms/AtlasProduction package.
        self.isJT=string.find(app.transform_archive,"JobTransform")
        if self.isJT>-1 and app.mode=="evgen":
            environment['T_CONTEXT'] = str(self.number_events_job) # needed to avoid prodsys failure mechanism based on a hardcoded minimum number of event of 5000 per job

            
        #       prepare job requirements
            
        if hasattr(job.backend,'requirements') and hasattr(job.backend.requirements,'sites') and hasattr(job.backend.requirements,'software') and hasattr(job.backend.requirements,'other') :
            requirements=job.backend.requirements
        else:
            requirements = AtlasLCGRequirements()
        
#        requirements.other.append('other.GlueCEStateStatus=="Production"') # missing production
        imax=string.rfind(app.atlas_rel,".")
        rel=string.atof(app.atlas_rel[:imax]) # to deal with string comparisons: [2-9].0.0 > 11.0.0. 
        if app.atlas_rel <= "11.4.0" or rel <=11.4:
            requirements.software=['VO-atlas-release-%s' % app.atlas_rel ]
        elif app.atlas_rel < "12.0.3":
            requirements.software=['VO-atlas-offline-%s' % app.atlas_rel ]
        elif app.atlas_rel >= "14.0.0" and app.atlas_rel<= "15.6.1":
            requirements.software=['VO-atlas-offline-%s-i686-slc4-gcc34-opt' % app.atlas_rel ]
        elif app.atlas_rel> "15.6.1":
            requirements.software=['VO-atlas-offline-%s-i686-slc5-gcc43-opt' % app.atlas_rel ]
        else:
            requirements.software=['VO-atlas-production-%s' % app.atlas_rel ]
        # case of prod_release set
        if app.prod_release:
            # no prod release tag before 13.0.X
            if app.atlas_rel < "14.0.0" and app.atlas_rel > "13.0.0":
                requirements.software=['VO-atlas-production-%s' % app.prod_release]
            elif app.atlas_rel>= "14.0.0" and app.atlas_rel<= "15.6.1":
                requirements.software=['VO-atlas-production-%s-i686-slc4-gcc34-opt' % app.prod_release]
            elif app.atlas_rel> "15.6.1":
                requirements.software=['VO-atlas-production-%s-i686-slc5-gcc43-opt' % app.prod_release]

        if app.transform_archive and string.find(app.transform_archive,"AtlasTier0")>-1:
            requirements.software=['VO-atlas-tier0-%s' % app.prod_release]
##        extraConfig=getConfig('defaults_AtlasLCGRequirements')
##        if  'dq2client_version' in extraConfig:
##            dq2client_version = extraConfig['dq2client_version']

##        if job.backend.requirements.dq2client_version:
##            dq2client_version = job.backend.requirements.dq2client_version
##        try:
##            assert dq2client_version!=""
##        except:
##            raise  ApplicationConfigurationError(None,"Please give a value to dq2client_version in job.backend.requirements.")
        
#        requirements.software += ['VO-atlas-dq2clients-%s' % dq2client_version]
#        requirements.other+=['RegExp("VO-atlas-dq2clients",other.GlueHostApplicationSoftwareRunTimeEnvironment)']

        # controlled relaxation for simple cases: one single input dataset, less than 200 subjobs. In this case, the subjobs can be submitted to the whole cloud.
        loosematch="true"
        if job.splitter and job.splitter.numsubjobs>200:
            loosematch="false"
        if job.inputdata and (job.inputdata.cavern or job.inputdata.minbias):
            loosematch="false"
# commented the nex block out as stage-in.sh can now ensure that the local copy is downloaded in the first attempt. However, as a safety net, we maintain the veto on complex jobs with pileup and or minbias, because they are heavy weight anyway and should not be run everywhere.
#        if app.dbrelease: 
#            loosematch="false"
        if len(job.backend.requirements.sites)>0:
            loosematch="false" # specified sites take precedence over cloud.
            
        userCloud=job.backend.requirements.cloud
        if userCloud=='ALL':
            userCloud='' # not supporting the AthenaLCGRequirements catch-all
        # By default: job to data, strict: target outsite and nothing else.
        requirements.sites=outsite
        
        if loosematch=="true" and userCloud :
            logger.debug("Your job qualifies for controlled relaxation of the current job-to-data policy. Now checking that requested cloud matches with input data")
            
            from dq2.info.TiersOfATLAS import whichCloud,ToACache
            targetSites=whichCloud(outsite)
            cloud=""
            for cloudID,sites in ToACache.dbcloud.iteritems():
                if sites==targetSites:
                    cloud=cloudID
            try:
                assert cloud==userCloud
            except:
                raise ApplicationConfigurationError(None,"Requested cloud: %s did not match selected processing cloud: %s. Reverting to submission to site %s" % (userCloud,cloud,outsite))

            requirements.cloud=cloud
            # looks like cloud has to be converted in a list of sites anyway, and this is not done in AtlasLCGRequirements.convert()... 
            allsites=requirements.list_sites_cloud()
            try:
                assert len(allsites)>0
            except:
                raise ApplicationConfigurationError(None,"Could not get any sites from the specified cloud: %s. You will have to specify a target site in job.backend.requirements.sites" % cloud)
            # need to weed out unwanted sites from excluded list
            excludedSites=requirements.excluded_sites
            goodsites=[]
            for checksite in allsites:
                selsite=True
                for site in excludedSites:
                    imax=site.find("_")
                    shortSite=site[:imax]
                    if shortSite in checksite:
                      #  print "site is excluded, skipping ", checksite
                        selsite=False
                        break
                if selsite and checksite not in goodsites:
                    goodsites.append(checksite)
#            print len(allsites),len(goodsites)
            if len(goodsites)>0:
                allsites=goodsites
            job.backend.requirements.sites=allsites
            logger.debug("Relaxing job to data policy to job to cloud. Selected cloud is %s" % cloud)

        logger.debug("master job submit?")
        
        if job.backend._name=="LCG" or job.backend._name=="Cronus" or job.backend._name=="Condor" or job.backend._name=="NG":
            return LCGJobConfig("",inputbox,[],outputbox,environment,[],requirements)
        else :
            return StandardJobConfig("",inputbox,[],outputbox,environment)



        
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        """Prepare the job"""

        inputbox=[ ]

        #       prepare environment
        environment={}
        environment=jobmasterconfig.env.copy()
        environment["INPUTDATASETS"]=""
        environment["INPUTFILES"]=""
        environment["INPUTTURLS"]=""

        alllfns=app.inputfiles+app.cavernfiles+app.mbfiles+app.dbfiles
        guids=app.turls
        guids.update(app.cavern_turls)
        guids.update(app.minbias_turls)
        guids.update(app.dbturls)
        
        infilenr=0
        for infile in alllfns:
            environment["INPUTFILES"]+="lfn[%d]='%s';" %(infilenr,infile)
            environment["INPUTDATASETS"]+="dset[%d]='%s';"%(infilenr,app.dsetmap[infile])
##            insites=app.sitemap[infile]
##            # compare with environment["OUTSITE"] and reorder if needed.
##            newinsites=self.sortSites(insites,environment["OUTSITE"])
##            environment["INPUTSITES"]+="site[%d]='%s';"%(infilenr,newinsites)
            environment["INPUTTURLS"]+="turl[%d]='%s';"%(infilenr,guids[infile])
            
            infilenr += 1


        logger.debug("%s %s %s" % (str(environment["INPUTDATASETS"]),str(environment["INPUTTURLS"]),str(environment["INPUTFILES"])))
        
        if environment["INPUTDATASETS"] :
            # Work around for glite WMS spaced environement variable problem
            inputbox += [ FileBuffer('inputdsets.conf',environment['INPUTDATASETS']+'\n') ]
        if environment["INPUTTURLS"] :
            # Work around for glite WMS spaced environement variable problem
            inputbox += [ FileBuffer('inputturls.conf',environment['INPUTTURLS']+'\n') ]
        if environment["INPUTFILES"] :
            # Work around for glite WMS spaced environement variable problem
            inputbox += [ FileBuffer('inputfiles.conf',environment['INPUTFILES']+'\n') ]

# now doing output files....
        job = app._getParent() # Returns job or subjob object

        outfilelist=""
        for type in app.outputpaths.keys():
            if type=="LOG" and "LOG" not in job.outputdata.outrootfiles:
                # logfiles are no longer saved in DQ2 datasets unless they are explicitly named in the outrootfiles dictionnary
                continue
            outfilelist+=app.outputpaths[type]+app.subjobsOutfiles[job.id][type]+" "

        environment["OUTPUTFILES"]=outfilelist
        # Work around for glite WMS spaced environement variable problem
        inputbox += [ FileBuffer('outputfiles.conf',environment['OUTPUTFILES']+'\n') ]        

 # setting up job wrapper arguments.       
        args=app.args
        trfargs=' '.join(app.args[4:])
        inputbox += [ FileBuffer('trfargs.conf',trfargs+'\n') ]
        jid=""
        if job._getRoot().subjobs:
            jid = job._getRoot().id
        else:
            jid = "%d" % job.id
        environment["OUTPUT_JOBID"]=str(jid) # used for versionning
        if app.dryrun:
            environment["DRYRUN"] = "TRUE"
        if app.dbrelease:
            environment["ATLASDBREL"]=app.dbrelease
        inputdata = []

        filename="wrapper.sh"
        exe = os.path.join(os.path.dirname(__file__),filename)

#       output sandbox
        outputbox =jobmasterconfig.outputbox

        if job.backend._name=="LCG" or job.backend._name=="Cronus" or job.backend._name=="Condor" or job.backend._name=="NG" or job.backend._name=="SGE":
            logger.debug("submission to %s" % job.backend._name)
            #       prepare job requirements
            requirements = jobmasterconfig.requirements
 
            if "INPUTTURLS" in environment:
                logger.debug(environment["INPUTTURLS"])
                if string.find(environment["INPUTTURLS"],"file:")>=0:
                    raise ApplicationConfigurationError(None,"Input file was found to be local, and LCG backend does not support replication of local files to the GRID yet. Please register your input dataset in DQ2 before resubmitting this job. Aborting")
            if string.lower(app.se_name)=="local":
                raise ApplicationConfigurationError(None,"Output file cannot be committed to local filesystem on a grid job. Please change se_name")

            lcg_job_config = LCGJobConfig(File(exe),inputbox,args,outputbox,environment,inputdata,requirements) 
            lcg_job_config.monitoring_svc = mc['AthenaMC/LCG']
            return lcg_job_config
        else:
            logger.debug("Backend %s not fully supported , will try our best anyway..." % job.backend._name)
            # if there are input data files and they are on the grid, prestage them on local area (use either app.datasets.input_dataset or /tmp/$login/data (and update environment["INPUTFILE"] accordingly inf the later is used...)
            # later development....

            return StandardJobConfig(File(exe),inputbox,args,outputbox,environment) 



allHandlers.add('AthenaMC','LCG',AthenaMCLCGRTHandler)
allHandlers.add('AthenaMC','Condor',AthenaMCLCGRTHandler)
allHandlers.add('AthenaMC','NG',AthenaMCLCGRTHandler)


config = getConfig('AthenaMC')
logger = getLogger('AthenaMC')
